/**
 * 
 */
package electric_vehicles.Agents;

import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.lang.Math;

import electric_vehicles.Enums.ChargeLocation;
import electric_vehicles.Enums.FuelType;
import electric_vehicles.Enums.InformationStrategy;
import electric_vehicles.Enums.MessageType;
import electric_vehicles.Enums.NeedWeight;
import electric_vehicles.Enums.SocialWeight;
import electric_vehicles.Media.Media;
import electric_vehicles.Media.MediaMessage;
import electric_vehicles.Memory.FuelTypeConcept;
import electric_vehicles.Memory.MemoryOthers;
import electric_vehicles.Memory.MemoryOwn;
import electric_vehicles.Memory.VehicleConcept;
import electric_vehicles.VehicleInfrastructure.Marketplace;
import electric_vehicles.VehicleInfrastructure.RechargeStation;
import electric_vehicles.VehicleInfrastructure.Tax;
import electric_vehicles.Vehicles.DriveLog;
import electric_vehicles.Vehicles.Vehicle;
import repast.simphony.context.Context;
import repast.simphony.engine.schedule.ScheduledMethod;
import repast.simphony.essentials.RepastEssentials;
import repast.simphony.random.RandomHelper;
import repast.simphony.util.ContextUtils;
import repast.simphony.util.collections.IndexedIterable;

//import java.util.Vector;

/**
 * @author Ayla
 *
 */
public class Agent 
{
	// *** GENERAL ***
	protected int d_id;
	
	// *** PERSONALITY ***
	protected Map<NeedWeight, Float> d_need_weights;
	protected float d_ambition;
	protected float d_uncertainty_tolerance;
	
	// *** NEEDS ***
	protected float d_satisfaction;
	protected float d_costs_satisfaction, d_functionality_satisfaction, d_social_satisfaction, d_preference_satisfaction;
	protected float d_uncertainty;	
	protected float d_costs_uncertainty, d_functionality_uncertainty, d_social_uncertainty, d_preference_uncertainty;
	// Costs
	protected float d_static_costs_weight;
	// Functionality
	protected int d_min_range;
	protected Map<ChargeLocation, Integer> d_max_charge_times;
	// Social
	protected Map<SocialWeight, Float> d_social_weights;
	protected Map<Integer, Float> d_friends;
	// Environmental
	protected float d_max_emissions;
	
	// *** ABILITIES ***
	protected boolean d_lease;
	protected boolean d_buys_new;
	protected float d_money;
	protected static float d_money_prop;
	protected boolean d_homecharge;
	protected boolean d_workcharge;

	// *** MEMORY ***
	protected MemoryOthers d_memoryothers;	// Store information about other agents
	protected MemoryOwn d_memoryown;		// Store all other information
	
	// *** OPTION ***
	protected Vehicle d_vehicle;	
	protected List<Integer> d_distances;

	// *** CHARACTERISTICS ***
	protected int d_income;
	protected int d_income_category;	
	protected int d_age;
	protected int d_location;
	protected float d_recharge_moment;
	protected List<Float> d_car_costs_range;
	protected Map<Integer, List<Integer>> d_distance_freq;
	protected int d_yearly_km;
	protected int d_yearly_km_category;
	
	/* 
	 * ******************************************************** CONSTRUCTOR *******************************************************
	 * 
	*/
	public Agent(int id, Map<NeedWeight,Float> need_weights, float ambition, float uncertainty_tolerance, float expertise,  Map<Integer, 
			List<Integer>> distance_freq, int min_range, Map<ChargeLocation, Integer> max_charge_times, Map<SocialWeight, Float> social_weights, float max_emissions, 
			float enjoy_innovation, int income_category, int income, int age, int location, boolean homecharge, float recharge_moment, FuelType current_fuel_type, int current_vehicle_age, 
			int vehicle_in_possesion_age, int yearly_km_category, int yearly_km, List<Float> car_costs_range, boolean buys_new, boolean lease)
	{		
		//  *** GENERAL ***
		d_id = id;
		
		// *** PERSONALITY ***
		d_need_weights = need_weights;
		d_ambition = ambition;
		d_uncertainty_tolerance = uncertainty_tolerance;
		
		// *** NEEDS ***
		d_satisfaction = 0;
		d_costs_satisfaction = d_functionality_satisfaction = d_social_satisfaction = d_preference_satisfaction = 0;
		d_uncertainty = 0;
		d_costs_uncertainty = d_functionality_uncertainty = d_social_uncertainty = d_preference_uncertainty = 0;
		// 	Costs
		d_static_costs_weight = (-(float)1/18) * yearly_km_category + ((float)29/36);
		// 	Functionality
		d_min_range = min_range;
		d_max_charge_times = max_charge_times;
		// 	Social
		d_social_weights = social_weights;
		d_friends = new HashMap<Integer, Float>();
		// 	Environmental
		d_max_emissions = max_emissions;
		
		// *** ABILITIES ***
		d_lease = lease;
		d_buys_new = buys_new;
		d_homecharge = homecharge;
		if (d_homecharge)
			RechargeStation.increaseHomeChargeAgentsCount();
		d_workcharge = false;
		d_money = 0;
		d_money_prop = 0.05f;

		// *** OPTION *** 
		d_vehicle = Marketplace.getInitialCar(current_fuel_type, yearly_km, current_vehicle_age, vehicle_in_possesion_age, car_costs_range, buys_new);
		d_distances = new LinkedList<Integer>();
		d_distance_freq = distance_freq;
		initializeDriveDistances();
				
		// *** MEMORY *** 
		d_memoryown = new MemoryOwn(d_vehicle, expertise);
		d_memoryothers = new MemoryOthers();
		
		// *** CHARACTERISTICS *** 
		d_income = income;
		d_income_category = income_category;
		d_age = age;
		d_location = location;
		d_recharge_moment = recharge_moment;
		d_car_costs_range = car_costs_range;		
		d_yearly_km = yearly_km;
		d_yearly_km_category = yearly_km_category;	
	}	
	
	/* 
	 * ****************************************************** STEP FUNCTION *******************************************************
	 * 
	*/
	@ScheduledMethod(start = 1 , interval = 1, priority = 2)
	public void step()
	{							
		// Drive 7 distances
		for (int count = 0; count < 7; ++count)
			drive(count);
		
		// Depending on satisfaction & uncertainty, seek new information on available cars
		performInformationSeekingStrategy(computeExperiencedNeedSatisfaction(), computeExperiencedNeedUncertainty());	
	
		// Purchase new vehicle?	
		decideToBuyNewVehicle();
	}
	
	/* 
	 * ***************************************************** DRIVING FUNCTIONS *****************************************************
	 * 
	*/		
	public void drive(int count)
	{				
		// Reset info at beginning of the tick ('week')
		if (count == 0)
		{
			// Did work place or home obtain charge station?
			checkChargeStations();	
			
			// At the start of each 'year', initialize an array containing all drive distances
			// for the coming 7 * 52 drives.
			if ((int)RepastEssentials.GetTickCount() % 52 == 0)
				initializeDriveDistances();
			
			d_memoryown.resetWeeklyDistance();
			d_memoryown.resetWeeklyEmissions();
			d_memoryown.resetFailedDrives();
			d_memoryown.resetLastStepFailureCosts();
		}
		// Randomly select a distance to drive
		int distance = selectDriveDistance();	
		
		// Is this a trip to work place?
		boolean workTrip = (d_vehicle.getFuelType() == FuelType.GASOLINE || !d_workcharge) ? false : isTripToWork(distance); 
				
		// Enough charge stations along the road?
		if (d_vehicle.getFuelType() == FuelType.ELECTRIC && !d_vehicle.driveSucceeds(distance, workTrip, d_recharge_moment))
		{
			d_memoryown.increaseFailedDrives();
			d_memoryown.updateExistenceNeeds(0, 0);
			return;
		}
		
		// Drive this distance and obtain log information
		DriveLog log = d_vehicle.getDriveLog(distance, workTrip, d_homecharge, d_recharge_moment);
		
		// Add this week's carbon emissions
		int emissions_distance = distance;
		if (d_vehicle.getFuelType() == FuelType.HYBRID)
			emissions_distance -= log.d_electricKmWithPHEV;
		d_memoryown.addWeeklyEmissions(emissions_distance * d_vehicle.getGasolineEmissions());
		d_memoryown.addWeeklyDistance(distance);

		
		// Determine and store satisfaction of this drive log
		determineDriveSatisfaction(log);
				
		// Store current aspects of the car in its vehicle concept (range, failure costs etc)
		updateOwnOpinionOfOwnVehicle();
	}
	
	public int selectDriveDistance()
	{			
	    return d_distances.remove(d_distances.size()-1);
	}
	
	public void determineDriveSatisfaction(DriveLog log)
	{		
		// ::::TAX COSTS::::
		float tax_satisfaction = 1;
		if (d_lease)
			tax_satisfaction = 1 - (float)Tax.getBijtellingLevel(d_vehicle.getEmissions()) / Tax.getMaxBijtellingLevel();
		else
			tax_satisfaction = 1 - (float)Tax.getTaxLevel(d_vehicle.getEmissions()) / Tax.getMaxTaxLevel();

		// ::::FAILURES::::
		//Determine any dissatisfaction from recent failures
		float failureCostsSatisfaction = d_memoryown.getFailureCostsSatisfaction();
		d_memoryown.addLastStepFailureCosts(log.d_failureCosts);
		// If a new failure occurred, reduce satisfaction related to failure costs
		if (log.d_failureCosts > 0)
			failureCostsSatisfaction = (float)1/log.d_failureCosts;
		// Otherwise, increase satisfaction because no failure occurred, with a max of 1
		else
			failureCostsSatisfaction = Math.min(failureCostsSatisfaction * 1.5f, 1);
		d_memoryown.setFailureCostsSatisfaction(failureCostsSatisfaction);
		
		// ::::ENERGY COSTS::::
		// If refueling took place, determine satisfaction with fuel costs
		if (log.d_refillCosts > 0)
		{
			float costsPerKm = log.d_refillCosts / log.d_refillEnergy;
			// Sigmoid function determines base satisfaction
			float energyCostsSatisfaction = (float) (0.5 - 0.5 * Math.tanh(2 * 10 * (costsPerKm - 0.1)));
			// Proportional change function determines whether temporary extra satisfaction drop/rise occurs
			energyCostsSatisfaction = Math.max(0, Math.min(1, energyCostsSatisfaction - (costsPerKm - d_memoryown.getLastCostsPerKm()) / d_memoryown.getLastCostsPerKm()));
			// Historical change function determines whether any previous temporary satisfaction drop/rise should be included
			float prev_temp_difference = (float) ((0.5 - 0.5 * Math.tanh(2 * 10 * (d_memoryown.getLastCostsPerKm() - 0.1))) - d_memoryown.getLastEnergyCostsSatisfaction());
			energyCostsSatisfaction = (float) Math.max(0, Math.min(1, energyCostsSatisfaction - (prev_temp_difference * 0.8)));
			d_memoryown.setLastCostsPerKm(costsPerKm);
			d_memoryown.setLastEnergyCostsSatisfaction(energyCostsSatisfaction);
		}
		// Else, use last satisfaction, but keep in mind a drop/rise for recent price fluctuations 
		else
		{
			float energyCostsSatisfaction = (float) (0.5 - 0.5 * Math.tanh(2 * 10 * (d_memoryown.getLastCostsPerKm() - 0.1)));
			// Historical change function determines whether any previous temporary satisfaction drop/rise should be included
			float prev_temp_difference = energyCostsSatisfaction - d_memoryown.getLastEnergyCostsSatisfaction();
			energyCostsSatisfaction = (float) Math.max(0, Math.min(1, energyCostsSatisfaction - (prev_temp_difference * 0.8)));
			d_memoryown.setLastEnergyCostsSatisfaction(energyCostsSatisfaction);
		}
		
		// ::::LATENESS::::
		// Determine (dis)satisfaction with lateness. Every 200km, agents are willing to be 15 minutes late
		float acceptable_delay = (float)log.d_distance / 200 * (float)d_max_charge_times.get(ChargeLocation.HIGHWAY);
		float timeSatisfaction = log.d_minutesLate == 0 ? 1 : 1 - Math.min(1, (float)log.d_minutesLate/acceptable_delay);
		
		// ::::FINAL SATISFACTION::::
		// Calculate average satisfaction. Income level weighs the importance of cost satisfaction
		// i.e. the richer, the less important the costs
		float static_costs_satisfaction = tax_satisfaction;
		float drive_costs_satisfaction = (d_memoryown.getLastEnergyCostsSatisfaction() + failureCostsSatisfaction) / 2;
		float total_costs_satisfaction = d_static_costs_weight * static_costs_satisfaction + (1 - d_static_costs_weight) * drive_costs_satisfaction;
		
		d_memoryown.updateExistenceNeeds(total_costs_satisfaction, timeSatisfaction);
	}
	
	public boolean isTripToWork(int distance)
	{
		if (distance > 10 && distance <= 50 && d_distance_freq.get(1).contains(50))
			return true;
		else if (distance > 50 && distance <= 100 && d_distance_freq.get(1).contains(100))
			return true;
		else if (distance > 100 && distance <= 200 && d_distance_freq.get(1).contains(200))
			return true;
		else
			return false;
	}
	
	/* 
	 * ************************************************ NEEDS SATISFACTION FUNCTIONS ***********************************************
	 * 
	*/	
	
	/* Compute satisfaction of currently owned vehicle
	 */
	public float computeExperiencedNeedSatisfaction()
	{
		//float existence_satisfaction = computeExistenceSatisfaction();
		
		d_costs_satisfaction = d_memoryown.getCostSatisfaction();
		d_functionality_satisfaction = d_memoryown.getFunctionalitySatisfaction();
		d_social_satisfaction = computeSocialSatisfaction(d_vehicle.getModel(), (d_costs_satisfaction + d_functionality_satisfaction)/2);
		d_preference_satisfaction = computePreferenceSatisfaction(d_vehicle.getModel());
				
		float satisfaction = d_costs_satisfaction * d_need_weights.get(NeedWeight.COSTS)
				+ d_functionality_satisfaction * d_need_weights.get(NeedWeight.FUNCTIONALITY)
				+ d_social_satisfaction * d_need_weights.get(NeedWeight.SOCIAL)
				+ d_preference_satisfaction * d_need_weights.get(NeedWeight.PREFERENCES);
		
		// ::::DISTANCE LIMITATION::::
		// Discount satisfaction whenever a distance could not be traveled during recent ticks
		if (d_vehicle.getFuelType() == FuelType.ELECTRIC && d_memoryown.getFailedDrives() > 0)
			satisfaction *= (1 - (d_memoryown.getFailedDrives()/7));	
		
		//if (d_buys_new)
		//	satisfaction *= (20 - Marketplace.getUpgradesSinceThisVehicle(d_vehicle.getModel())) / 20;
		
		// Store for statistics
		d_satisfaction = satisfaction;
		
		return satisfaction;
	}
	
	/* Compute satisfaction of any car that is not currently owned
	 */
	public float computePerceivedNeedSatisfaction(String model, float costs_satisfaction)
	{
		//float existence_satisfaction = computeExistenceSatisfaction(model, costs_satisfaction);
		float functionality_satisfaction = computeFunctionalitySatisfaction(model);
		float social_satisfaction = computeSocialSatisfaction(model, (functionality_satisfaction + costs_satisfaction)/2);
		float preference_satisfaction = computePreferenceSatisfaction(model);
						
		float satisfaction = costs_satisfaction * d_need_weights.get(NeedWeight.COSTS)
				+ functionality_satisfaction * d_need_weights.get(NeedWeight.FUNCTIONALITY)
				+ social_satisfaction * d_need_weights.get(NeedWeight.SOCIAL)
				+ preference_satisfaction * d_need_weights.get(NeedWeight.PREFERENCES);
		
		VehicleConcept vehicle_concept = this.getVehicleConcepts().get(model);
		if (vehicle_concept.getFuelType() == FuelType.ELECTRIC && RechargeStation.getChargeProbability(ChargeLocation.HIGHWAY) < 1)
		{
			float recharge_prob = RechargeStation.getChargeProbability(ChargeLocation.HIGHWAY);			
			// Look at each frequency (e.g. '1 once a day', '2 once a month') and corresponding ranges (e.g. '100 km', '500 km')
			for (Map.Entry<Integer, List<Integer>> entry : d_distance_freq.entrySet())
			{
				for (int value : entry.getValue())
				{
					// If one of the ranges is larger than the range of the vehicle, reduce satisfaction
					// depending on frequency (key) with which this range (value) is driven. I.e. more frequent
					// distances result in higher penalties.
					if (vehicle_concept.getRange() < value)
					{
						switch(entry.getKey())
						{
							case 1: 	satisfaction -= (satisfaction * 0.4 * (1 - recharge_prob));
							case 2: 	satisfaction -= (satisfaction * 0.2 * (1 - recharge_prob));
							case 3: 	satisfaction -= (satisfaction * 0.1 * (1 - recharge_prob));
							default: 	;
						}
					}
				}
			}
		}		
		
		return satisfaction;
	}
	
	public float computeYearlyCosts(String model, boolean totalLoss)
	{
		// See children classes
		System.out.println("ERROR: Parent function computeYearlyCosts(String model) is called.");
		return 0.0f;
	}
	
	public float computeFunctionalitySatisfaction(String model)
	{
		VehicleConcept vehicle_concept = this.getVehicleConcepts().get(model);
		FuelTypeConcept fueltype_concept =  this.getFuelTypeConcepts().get(vehicle_concept.getFuelType());
		
		// If new vehicle's fuel type is the same as current fuel type, use current functionality satisfaction
		// note: phevs provide the same functionality (range) as gasoline vehicles
		if (vehicle_concept.getFuelType() != FuelType.ELECTRIC && d_vehicle.getFuelType() != FuelType.ELECTRIC)
			return d_memoryown.getFunctionalitySatisfaction();
		else if (vehicle_concept.getFuelType() == FuelType.ELECTRIC && d_vehicle.getFuelType() == FuelType.ELECTRIC)
		{
			// If new vehicle is an EV AND its range deviates from current range (> 10%), adapt current functionality satisfaction
			if (vehicle_concept.getFuelType() == FuelType.ELECTRIC && (vehicle_concept.getRange() < d_vehicle.getRange()*0.9 || vehicle_concept.getRange() > d_vehicle.getRange()*1.1))
				return (vehicle_concept.getRange()/d_vehicle.getRange()) * d_memoryown.getFunctionalitySatisfaction();
			return d_memoryown.getFunctionalitySatisfaction();
		}
		
		// ::::RANGE::::
		float range_satisfaction = Math.min(1, vehicle_concept.getRange() / this.getMinRange());
		
		// ::::CHARGE TIME::::
		float charge_time_slow = fueltype_concept.getChargeTime();
		float charge_time_fast = fueltype_concept.getChargeTimeFast();		
		List<Float> locations = new LinkedList<Float>();
		// For each location, does agent uses this location to charge (d_max_charge_times.get(ChargeLocation.HOME) != 0)? If so, determine satisfaction
		if (d_max_charge_times.get(ChargeLocation.HOME) != 0 && vehicle_concept.getFuelType() == FuelType.ELECTRIC)
		{
			if (d_homecharge)
				locations.add(Math.max(0, 1 - (charge_time_slow / (float)d_max_charge_times.get(ChargeLocation.HOME))));
			else 									 // Agent only drives electric and cannot charge at home. Very unsatisfying.
				locations.add(0.0f);
		}
		if (d_max_charge_times.get(ChargeLocation.WORK) != 0 && vehicle_concept.getFuelType() == FuelType.ELECTRIC)
		{
			if (d_workcharge)
				locations.add(Math.max(0, 1 - (charge_time_slow / (float)d_max_charge_times.get(ChargeLocation.WORK))));
			else 									// Agent only drives electric and cannot charge at work. Very unsatisfying.
				locations.add(0.0f);
		}
		if (d_max_charge_times.get(ChargeLocation.HIGHWAY) != 0)
			locations.add(Math.max(0, 1 - (charge_time_fast / (float)d_max_charge_times.get(ChargeLocation.HIGHWAY))));
		float total_satisfaction = 0;
		for (float location : locations)
			total_satisfaction += location;
		float charge_time_satisfaction = locations.size() == 0 ? 0 : total_satisfaction / locations.size();
				
		float functionality_satisfaction = (range_satisfaction + charge_time_satisfaction) / 2;

		return functionality_satisfaction;
	}
	
	
	/* Compute social satisfaction for any car. Existence satisfaction is passed because
	 * this same method is used for currently owned and currently not owned cars.
	 */
	public float computeSocialSatisfaction(String model, float existence_satisfaction)
	{
		VehicleConcept vehicle_concept = this.getVehicleConcepts().get(model);
		
		Map<SocialWeight, Float> stats = d_memoryothers.getSocialSatisfactionFactors(vehicle_concept, existence_satisfaction);
		
		float conformity = stats.get(SocialWeight.CONFORMITY);
		float anti_conformity = stats.get(SocialWeight.ANTI_CONFORMITY);
		float superiority = stats.get(SocialWeight.SUPERIORITY);
		
		float social_satisfaction = conformity * d_social_weights.get(SocialWeight.CONFORMITY)
				+ anti_conformity * d_social_weights.get(SocialWeight.ANTI_CONFORMITY)
				+ superiority * d_social_weights.get(SocialWeight.SUPERIORITY); 
						
		return social_satisfaction;
	}
	
	/* Compute preference satisfaction for any car 
	 * (same method is used for currently owned and currently not owned cars)
	 */
	public float computePreferenceSatisfaction(String model)
	{
		float env_satisfaction;
		float emissions = model == d_vehicle.getModel() ? d_vehicle.getEmissions() : getVehicleConcepts().get(model).getEmissions();
		env_satisfaction = emissions == 0 ? 1 : 1 - Math.min(1,  emissions / d_max_emissions);

		return env_satisfaction;
	}
	
	/* 
	 * ************************************************ NEEDS UNCERTAINTY FUNCTIONS ************************************************
	 * 
	*/	
	
	/* Compute uncertainty of currently owned car
	 */
	public float computeExperiencedNeedUncertainty()
	{
		VehicleConcept vehicle_concept = this.getVehicleConcepts().get(d_vehicle.getModel());
		
		d_costs_uncertainty = d_memoryown.getCostUncertainty();		
		d_functionality_uncertainty = d_memoryown.getFunctionalityUncertainty();
		d_social_uncertainty = d_memoryothers.getSocialUncertainty(vehicle_concept.getFuelType());
		d_preference_uncertainty = vehicle_concept.getPreferenceUncertainty();

		float uncertainty = d_costs_uncertainty * d_need_weights.get(NeedWeight.COSTS) +
				d_functionality_uncertainty * d_need_weights.get(NeedWeight.FUNCTIONALITY) +
				d_social_uncertainty * d_need_weights.get(NeedWeight.SOCIAL) +
				d_preference_uncertainty* d_need_weights.get(NeedWeight.PREFERENCES) ;
		
		// Store for statistics
		d_uncertainty = uncertainty;
		
		return uncertainty;
	}	
	
	/* Compute uncertainty of any car that is not currently owned
	 */
	public float computePerceivedNeedUncertainty(String model)
	{					
		float cost_uncertainty = computePerceivedCostUncertainty(model);
		float functionality_uncertainty = computePerceivedFuncUncertainty(model);
		float social_uncertainty = computePerceivedSocialUncertainty(model);
		float preference_uncertainty = computePerceivedPreferenceUncertainty(model);

		float uncertainty = cost_uncertainty * d_need_weights.get(NeedWeight.COSTS) +
				functionality_uncertainty * d_need_weights.get(NeedWeight.FUNCTIONALITY) +
				social_uncertainty * d_need_weights.get(NeedWeight.SOCIAL) +
				preference_uncertainty* d_need_weights.get(NeedWeight.PREFERENCES) ;

		return uncertainty;
	}	
	

	public float computePerceivedCostUncertainty(String model)
	{
		VehicleConcept vc = this.getVehicleConcepts().get(model);
		FuelTypeConcept ft =  this.getFuelTypeConcepts().get(vc.getFuelType());

		return (vc.getPriceUncertainty() + ft.getKmCostsUncertainty() + ft.getFailureCostsUncertainty() + vc.getTaxUncertainty(d_lease)) / 4;	
	}
	
	public float computePerceivedFuncUncertainty(String model)
	{
		VehicleConcept vc = this.getVehicleConcepts().get(model);
		FuelTypeConcept ft =  this.getFuelTypeConcepts().get(vc.getFuelType());
		return (vc.getRangeUncertainty() + ft.getChargeTimeUncertainty())/2;
	}
	
	public float computePerceivedSocialUncertainty(String model)
	{
		return d_memoryothers.getSocialUncertainty(this.getVehicleConcepts().get(model).getFuelType());
	}
	
	public float computePerceivedPreferenceUncertainty(String model)
	{
		return this.getVehicleConcepts().get(model).getPreferenceUncertainty();
	}
	

	/* 
	 * ************************************************ DECISION STRATEGY FUNCTIONS ************************************************
	 * 
	*/	
	
	void performInformationSeekingStrategy(float satisfaction, float uncertainty)
	{
		// Depending on level of satisfaction/uncertainty and ambition/uncertainty_tolerance, determine whether agent acts on its satisfaction and/or uncertainty.
		boolean satisfied = RandomHelper.nextDoubleFromTo(0, 1) <= Math.pow(satisfaction, d_ambition) ? true : false;
		boolean uncertain = RandomHelper.nextDoubleFromTo(0, 1) <= Math.pow(uncertainty, d_uncertainty_tolerance) ? true : false;
		
		// Keep track of how often agent is (un)satisfied
		d_memoryown.addSatisfaction(satisfied);
		
		// Select information seeking behaviour
		if (satisfied && !uncertain)
		{
			d_memoryown.setLastStrategy(InformationStrategy.REPETITION);
			return;// Repetition, do nothing
		}
		else if (satisfied && uncertain)
			imitate();// Imitation
		else if (!satisfied && uncertain)
			inquire();// Inquiring
		else if (!satisfied && !uncertain)
			optimise();// Optimizing		
	}
	
	// Discuss personal vehicles with close friends
	void imitate()
	{
		d_memoryown.setLastStrategy(InformationStrategy.IMITATION);
		communicate(false);
	}
	
	// Discuss most known vehicles with less close friends
	void inquire()
	{
		d_memoryown.setLastStrategy(InformationStrategy.INQUIRING);
		communicate(true);
	}
	
	// Seek information in media
	void optimise()
	{
		d_memoryown.setLastStrategy(InformationStrategy.OPTIMISING);
		List<MediaMessage> messages = Media.getAdvertisements();	// Get advertisements
		increaseExpertise(messages.size());							// Update expertise of this agent
		for (MediaMessage message : messages)						// Update beliefs
			updateMediaOpinion(message);		

	}
	/* 
	 * ********************************************* NETWORK & COMMUNICATION FUNCTIONS *********************************************
	 * 
	*/
	
	/* With a random probability, exchange information with other agents
	 * in the simulation. 
	 */
	public void communicate(boolean inquire)
	{
		int threshold = inquire ? 4 : 2;
		
		// Obtain iterator over all agents in the simulation
		@SuppressWarnings("unchecked")
		Context<Agent> context = (Context<Agent>)ContextUtils.getContext(this);			
		IndexedIterable<Agent> agents = (IndexedIterable<Agent>)context.getObjects(Agent.class);
		
		// Access each friend
		for (Map.Entry<Integer, Float> entry : d_friends.entrySet())
		{
			Agent agent = agents.get(entry.getKey());
						
			// Randomly decide whether to communicate, depending on similarity between agents
			float similarity = entry.getValue();
			float random = (float) RandomHelper.nextDoubleFromTo(0, 1);
						
			if (random < Math.pow(similarity, threshold))
				exchangeInformation(agent, similarity, inquire);
		}
	}
	
	public void exchangeInformation(Agent other, float similarity,  boolean inquire)
	{
		// Always ask about other agent's current car and fuel type and tell about own current car and fuel type
		List<VehicleConcept> others_vehicles_opinion = new LinkedList<VehicleConcept>();
		List<VehicleConcept> own_vehicles_opinion = new LinkedList<VehicleConcept>();
		others_vehicles_opinion.add(other.getVehicleConcepts().get(other.getVehicle().getModel()).copy());
		own_vehicles_opinion.add(this.getVehicleConcepts().get(d_vehicle.getModel()).copy());
		
		Map<FuelType, FuelTypeConcept> others_fueltype_opinion = new HashMap<FuelType, FuelTypeConcept>();
		Map<FuelType, FuelTypeConcept> own_fueltype_opinion = new HashMap<FuelType, FuelTypeConcept>();
		others_fueltype_opinion.put(other.getVehicle().getFuelType(), other.getFuelTypeConcepts().get(other.getVehicle().getFuelType()).copy());
		own_fueltype_opinion.put(d_vehicle.getFuelType(), this.getFuelTypeConcepts().get(d_vehicle.getFuelType()).copy());
		
		// If inquiring, talk about other cars as well
		if (inquire)
		{
			// With probability of 0.5 discuss each car the other agent knows
			List<VehicleConcept> concepts_other = new LinkedList<VehicleConcept>(other.getVehicleConcepts().values());
			Collections.shuffle(concepts_other);
			for (VehicleConcept vehicle_concept : concepts_other)
			{
				if (vehicle_concept.getModel() != other.getVehicle().getModel() && RandomHelper.nextDoubleFromTo(0, 1) <= 0.5 )
				{
					others_vehicles_opinion.add(vehicle_concept.copy());
					if (!others_fueltype_opinion.containsKey(vehicle_concept.getFuelType()))
						others_fueltype_opinion.put(vehicle_concept.getFuelType(), other.getFuelTypeConcepts().get(vehicle_concept.getFuelType()).copy());
				}
			}
			
			// With probability of 0.5 discuss each car this agent knows
			List<VehicleConcept> concepts_own = new LinkedList<VehicleConcept>(this.getVehicleConcepts().values());
			Collections.shuffle(concepts_own);
			for (VehicleConcept vehicle_concept : concepts_own)
			{
				if (vehicle_concept.getModel() != d_vehicle.getModel() && RandomHelper.nextDoubleFromTo(0, 1) <= 0.5)
				{
					own_vehicles_opinion.add(vehicle_concept.copy());
					if (!own_fueltype_opinion.containsKey(vehicle_concept.getFuelType()))
						own_fueltype_opinion.put(vehicle_concept.getFuelType(), this.getFuelTypeConcepts().get(vehicle_concept.getFuelType()).copy());
				}
			}
			// Update expertise of both agents
			this.increaseExpertise(others_vehicles_opinion.size());
			other.increaseExpertise(own_vehicles_opinion.size());
		}
		
		// Update info on agent in memory
		this.addMemoryOthers(other.getId(), others_vehicles_opinion.get(0), other.getExistenceSatisfaction(), other.getInformationCount(), similarity); 
		other.addMemoryOthers(d_id, own_vehicles_opinion.get(0), d_memoryown.getExistenceSatisfaction(), this.getInformationCount(), similarity);
		
		// Update info on vehicles in memory
		this.updateOthersOpinionOnVehicles(others_vehicles_opinion, other.getInformationCount());
		other.updateOthersOpinionOnVehicles(own_vehicles_opinion, this.getInformationCount());
		
		// Update info on fueltypes in memory
		this.updateOthersOpinionOnFuelTypes(others_fueltype_opinion, other.getInformationCount());
		other.updateOthersOpinionOnFuelTypes(own_fueltype_opinion, this.getInformationCount());
	}
	
	/* Computes the similarity between this agent and another. Higher similarity increases
	 * the likelihood that two agents will communicate. 
	 */	
	public float determineSimilarity(Agent other)
	{
		// Income similarity (don't use when one of the participants answered 'unknown')
		float income_similarity;
		income_similarity = 1 - (float)Math.abs(other.getIncome() - d_income_category)/10;	
		
		// Need similarity
		float need_similarity = Math.max(0, 1 - Math.abs(other.getNeedWeights().get(NeedWeight.COSTS) - d_need_weights.get(NeedWeight.COSTS)) -
									Math.abs(other.getNeedWeights().get(NeedWeight.FUNCTIONALITY) - d_need_weights.get(NeedWeight.FUNCTIONALITY)) -
									Math.abs(other.getNeedWeights().get(NeedWeight.SOCIAL) - d_need_weights.get(NeedWeight.SOCIAL)) -
									Math.abs(other.getNeedWeights().get(NeedWeight.PREFERENCES) - d_need_weights.get(NeedWeight.PREFERENCES)));

		// Ambition similarity
		float ambition_similarity = 1 - Math.abs(other.getAmbition() - d_ambition);
		
		// Behaviour similarity
		float behaviour_similarity = 1 - (float)Math.abs(other.getKmPerYear() - d_yearly_km)/10;
		
		// Age similarity
		float age_similarity = (float) (Math.abs(other.getAge() - d_age) > 40 ? 0 : Math.max(0, Math.pow(1 - (float)(Math.abs(other.getAge() - d_age))/40, 2)));
				
		// Distance similarity
		float location_similarity = other.getLocation() == d_location ? 1 : 0;
				
		// Final similarity
		return (income_similarity + need_similarity + ambition_similarity + behaviour_similarity + age_similarity + location_similarity)/6;
	}
	
	/* 
	 * ************************************************** UPDATE MEMORY FUNCTIONS ***************************************************
	 * 
	*/	

	public void updateOthersOpinionOnVehicles(List<VehicleConcept> vehicles_opinion, int informationCount)
	{				
		float informative_weight = Math.max(0.01f, (float)informationCount/d_memoryothers.getMaxInformationCount());
		
		for (VehicleConcept vehicle_concept : vehicles_opinion)
		{				
			// In case this car model is currently owned, don't update beliefs
			if (vehicle_concept.getModel().equals(d_vehicle.getModel()))
				continue;
			// Otherwise, update all stats or create a new vehicle concept if model is unknown.
			else if (d_memoryown.getVehicleConcepts().containsKey(vehicle_concept.getModel()))
				getVehicleConcept(vehicle_concept.getModel()).updateStats(vehicle_concept, informative_weight);
			else
				getVehicleConcepts().put(vehicle_concept.getModel(), vehicle_concept);
		}
	}
	
	public void updateOthersOpinionOnFuelTypes(Map<FuelType, FuelTypeConcept> fueltype_opinion, int informationCount)
	{				
		float informative_weight = (float)informationCount/d_memoryothers.getMaxInformationCount();
		
		for (FuelTypeConcept fueltype_concept : fueltype_opinion.values())
		{				
			// In case this car model is currently owned, don't update beliefs
			if (fueltype_concept.getFuelType().equals(d_vehicle.getFuelType()))
				continue;
			// Otherwise, update all stats 
			else
				getFuelTypeConcept(fueltype_concept.getFuelType()).updateStats(fueltype_concept, informative_weight);		
		}
	}
	
	public void updateMediaOpinion(MediaMessage message)
	{
		float media_weight = 1;
				
		if (message.messageType == MessageType.MEDIA_MESSAGE_VEHICLE)
		{
			// In case this car model is currently owned, don't update beliefs
			if (message.model.equals(d_vehicle.getModel()))
				return;
			// Otherwise, update all stats or create a new vehicle concept if model is unknown.
			else if (d_memoryown.getVehicleConcepts().containsKey(message.model))
				getVehicleConcept(message.model).updateStats(message.asVehicleConcept(), media_weight);
				
			else
				getVehicleConcepts().put(message.model, message.asVehicleConcept());
		}
		else if (message.messageType == MessageType.MEDIA_MESSAGE_FUEL)
		{
			// In case this fuel type is currently owned, don't update beliefs
			if (message.fuelType.equals(d_vehicle.getFuelType()))
				return;
			// Otherwise, update all stats or create a new fuel type concept if fuel type is unknown.
			else 
				getFuelTypeConcept(message.fuelType).updateStats(message.asFuelTypeConcept(), media_weight);
		}
	}
	
	public void updateOwnOpinionOfOwnVehicle()
	{
		d_memoryown.getVehicleConcepts().get(d_vehicle.getModel()).updateStatsOwnVehicle(d_vehicle);
		d_memoryown.getFuelTypeConcepts().get(d_vehicle.getFuelType()).updateStatsOwnVehicle(d_vehicle);
	}
	
	/* 
	 * ***************************************************** NEW CAR FUNCTIONS ******************************************************
	 * 
	*/	
	
	public boolean modelWithinPriceRange(String model)
	{
		// See children classes
		System.out.println("ERROR: Parent function modelWithinPriceRange(String model) is called.");
		return true;
	}
	
	public void decideToBuyNewVehicle()
	{		
		// See children classes
		System.out.println("ERROR: Parent function decideToBuyNewVehicle(String model) is called.");
	}
	
	
	/* 
	 * ******************************************************* UTIL FUNCTIONS ******************************************************
	 * 
	*/
	public int getId()										{ return d_id; }	
	
	// Personality
	public Map<NeedWeight,Float> getNeedWeights()			{ return d_need_weights; }
	public float getAmbition()								{ return d_ambition; }
	public float getUncertaintyTolerance()					{ return d_uncertainty_tolerance; }
	
	// Existence need
	public Map<Integer,List<Integer>> getDistanceFreq()		{ return d_distance_freq; }
	public int getMinRange()								{ return d_min_range; }
	public Map<ChargeLocation, Integer> getMaxChargeTime()	{ return d_max_charge_times; }
	
	// Social need
	public Map<SocialWeight, Float> getSocialWeights()		{ return d_social_weights; }
	
	// Preference need
	public float getMaxEmissions()							{ return d_max_emissions; }
	
	// Abilities
	public static float getMoneyProp()						{ return d_money_prop; }
	public static void adjustMoneyProp(float adjustment)	{ d_money_prop += adjustment; }
	public int getIncome()									{ return d_income_category; }
	public int getAge()										{ return d_age; }
	public int getLocation()								{ return d_location; }
	public int getKmPerYear()								{ return d_yearly_km; }
	
	// Memory
	public int getInformationCount()							{ return d_memoryothers.getInformationCount(); }
	public int getMaxInformationCount()							{ return d_memoryothers.getMaxInformationCount(); }
	public float getExistenceSatisfaction()						{ return d_memoryown.getExistenceSatisfaction(); }
	public Map<String, VehicleConcept> getVehicleConcepts()		{ return d_memoryown.getVehicleConcepts(); }
	public Map<FuelType, FuelTypeConcept> getFuelTypeConcepts()	{ return d_memoryown.getFuelTypeConcepts(); }
	public VehicleConcept getVehicleConcept(String model)		{ return d_memoryown.getVehicleConcepts().get(model); }
	public FuelTypeConcept getFuelTypeConcept(FuelType type)	{ return d_memoryown.getFuelTypeConcepts().get(type); }
	public void increaseExpertise(int incr)						{ d_memoryothers.increaseInformationCount(incr); }
	public void addMemoryOthers(int agent_id, VehicleConcept vehicle, float satisfaction, int information_count, float normative)
	{
		d_memoryothers.addMemory(agent_id, vehicle, satisfaction, information_count, normative);
	}
	
	// Option
	public Vehicle getVehicle()							{ return d_vehicle; }
	
	// Members used for statistics
	public int hasEV()									{ return d_vehicle.getFuelType() == FuelType.ELECTRIC ? 1 : 0; }
	public int hasFuel()								{ return d_vehicle.getFuelType() == FuelType.GASOLINE ? 1 : 0; }
	public int hasHybrid()								{ return d_vehicle.getFuelType() == FuelType.HYBRID ? 1 : 0; }
	public FuelType getFuelType()						{ return d_vehicle.getFuelType(); }
	public int inDebt()									{ return d_money < 0 ? 1 : 0; }
	public boolean hasHomeCharge()						{ return d_homecharge; }
	public boolean hasWorkCharge()						{ return d_workcharge; }
	public float vehicleAge()							{ return (float) ((RepastEssentials.GetTickCount() - d_vehicle.getFirstPurchaseTick())/52); }
	public int vehicleAKmCount()						{ return d_vehicle.getKmCount(); }
	public float averageOwnershipTime()					{ return (float) ( RepastEssentials.GetTickCount() - d_vehicle.getPurchaseTick() ) / 52; }
	public boolean canBuyModel(String model)			{ return getVehicleConcepts().containsKey(model) && modelWithinPriceRange(model); }
	public boolean isLeaser()							{ return d_lease; }
	public boolean buysNew()							{ return d_buys_new; }
	public int getWeeklyDistance()						{ return d_memoryown.getWeeklyDistance(); }
	public int getWeeklyEmissions()						{ return d_memoryown.getWeeklyEmissions(); }
	public InformationStrategy getLastStrategy()		{ return d_memoryown.getLastStrategy(); }
	//Experiences
	public float getSatisfaction()						{ return d_satisfaction; }
	public float getCostsSatisfaction()					{ return d_costs_satisfaction; }
	public float getFunctionalitySatisfaction()			{ return d_functionality_satisfaction; }
	public float getSocialSatisfaction()				{ return d_social_satisfaction; }
	public float getPreferenceSatisfaction()			{ return d_preference_satisfaction; }
	public float getUncertainty()						{ return d_uncertainty; }
	public float getCostsUncertainty()					{ return d_costs_uncertainty; }
	public float getFunctionalityUncertainty()			{ return d_functionality_uncertainty; }
	public float getSocialUncertainty()					{ return d_social_uncertainty; }
	public float getPreferenceUncertainty()				{ return d_preference_uncertainty; }
	
	public int satisfiedWithCurrentVehicle()			{ return d_memoryown.satisfiedWithCurrentVehicle()? 1 : 0; }
	
	/* 
	 * *************************************************** INITIALISE FUNCTIONS ****************************************************
	 * 
	*/	
	
	public int getDistanceFromKMRange(int kmRange)
	{
		switch(kmRange)
		{
		case 0: 	return 0;
		case 10:	return RandomHelper.nextIntFromTo(1, 10);
		case 50:	return RandomHelper.nextIntFromTo(11, 50);
		case 100:	return RandomHelper.nextIntFromTo(51, 100);
		case 200:	return RandomHelper.nextIntFromTo(101, 200);
		case 500:	return RandomHelper.nextIntFromTo(201, 500);
		case 1000:	return RandomHelper.nextIntFromTo(501, 1500);
		default:	System.out.println("WARNING: INVALID KM RANGE PASSED TO getDistanceFromKMRange() FUNCTION");
					return 0;
		}
	}
	
	@ScheduledMethod(start = 0, priority = 3)
	public void initializeFriendships()
	{
		// Obtain iterator over all agents
		@SuppressWarnings("unchecked")
		Context<Agent> context = (Context<Agent>)ContextUtils.getContext(this);			
		IndexedIterable<Agent> agents = (IndexedIterable<Agent>)context.getObjects(Agent.class);
		
		Map<Integer, Float> friends = new TreeMap<Integer, Float>();
		
		// Access each agent
		for(int i = 0; i < agents.size() ; ++i)
		{
			Agent agent = agents.get(i);
						
			// Then add and remove friends based on similarity measure
			if (agent.getId() != d_id  && RandomHelper.nextDoubleFromTo(0, 1) <= 0.2)
			{
				float similarity = determineSimilarity(agent);				
				float lowest_similarity = friends.size() == 0 ? 0 : Collections.min(friends.values());
				
				// First random 15 agents are always added
				if (friends.size() < 15)
				{
					friends.put(i, similarity);
					continue;
				}
				
				// Remaining agents are only added if they are more similar
				else if (similarity > lowest_similarity)
				{
					// Remove least similar agent
					for (Map.Entry<Integer, Float> set: friends.entrySet())
					{
						if (set.getValue() == lowest_similarity)
						{
							friends.remove(set.getKey());
							friends.put(i, similarity);	
							break;
						}
					}
				}
			}
		}		
		// Continue adding friends, until 15 OR all agents, except for itself, are added.
		while(friends.size() < 15 && agents.size() - 1 > friends.size())
		{
			int random_agent = RandomHelper.nextIntFromTo(0, agents.size() - 1);
			if (!friends.containsKey(random_agent) && agents.get(random_agent).getId() != d_id )
				friends.put(random_agent, determineSimilarity(agents.get(random_agent)));
		}
		
		d_friends = friends;
	}
	
	public void initializeDriveDistances()
	{	
		List<Integer> distances = new LinkedList<Integer>();

		int total_days = 0;
		
		// Select distances that are driven at least once a year
		List<Integer> ranges = d_distance_freq.get(3);
	    for (int range : ranges)
	    {
	    	int number_of_days = RandomHelper.nextIntFromTo(1, 11);
	    	for (int day = 0; day < number_of_days; ++day)
	    		distances.add(getDistanceFromKMRange(range));
	    	total_days += number_of_days;
	    }
	    
		// Select distances that are driven at least once a month
	    ranges = d_distance_freq.get(2);
	    for (int range : ranges)
	    {
	    	int number_of_days = RandomHelper.nextIntFromTo(12, 51);
	    	for (int day = 0; day < number_of_days; ++day)
	    		distances.add(getDistanceFromKMRange(range));
	    	total_days += number_of_days;
	    }
	    
	    // Distribute remaining days of year over all ranges that are driven
	    // at least once a week, while making sure each range is driven at
	    // least on 52 days.
	    ranges = d_distance_freq.get(1);
	    int days_left = 365 - total_days - ranges.size()*52;
	    for (int i = 0; i < ranges.size(); ++i)
	    {
	    	int number_of_days;
	    	if (i == ranges.size() - 1)
	    		number_of_days = days_left;
	    	else
	    	{
	    		number_of_days = days_left > 0 ? RandomHelper.nextIntFromTo(1, days_left) : 0;
		    	days_left -= number_of_days;
	    	}	    	
	    	if (number_of_days < 0 || number_of_days > 313)
	    	{
	    		System.out.println("Number of daily drives in range " + Integer.toString(ranges.get(i)) + ": " + Integer.toString(number_of_days + 52));
	    		System.out.println("Way too many days: " + number_of_days + " total days: " + total_days + " days left: " + days_left + " table: " + d_distance_freq);
	    	}
	    		
	    	for (int day = 0; day < 52 + number_of_days; ++day)
	    		distances.add(getDistanceFromKMRange(ranges.get(i)));
	    }
	    
	    if (distances.size() != 365)
	    	System.out.println("ERROR: Drive distance table contains " + distances.size() + " distances, not 365.");
	    
	    Collections.shuffle(distances);
	    	    
	    d_distances = distances;
	}
	
	public void checkChargeStations()
	{
		// Obtain number of agents
		@SuppressWarnings("unchecked")
		int agents_count = ContextUtils.getContext(this).getObjects(Agent.class).size();	
		
		if (!d_workcharge && (float)RechargeStation.getWorkChargeAgentsCount() / agents_count < RechargeStation.getChargeProbability(ChargeLocation.WORK))
		{
			d_workcharge = true;
			RechargeStation.increaseWorkChargeAgentsCount();
		}
		if (!d_homecharge && (float)RechargeStation.getHomeChargeAgentsCount() / agents_count < RechargeStation.getChargeProbability(ChargeLocation.HOME))
		{
			d_homecharge = true;
			RechargeStation.increaseHomeChargeAgentsCount();
		}
	}
	
}

