/**
 * 
 */
package electric_vehicles.Agents;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import repast.simphony.essentials.RepastEssentials;
import electric_vehicles.Enums.ChargeLocation;
import electric_vehicles.Enums.FuelType;
import electric_vehicles.Enums.NeedWeight;
import electric_vehicles.Enums.SocialWeight;
import electric_vehicles.Memory.FuelTypeConcept;
import electric_vehicles.Memory.VehicleConcept;
import electric_vehicles.VehicleInfrastructure.Marketplace;
import electric_vehicles.VehicleInfrastructure.Tax;
import electric_vehicles.Vehicles.Vehicle;

/**
 * @author Ayla
 *
 */
public class AgentLease extends Agent {
	
	
	public AgentLease(int id, Map<NeedWeight,Float> need_weights, float ambition, float uncertainty_tolerance, float expertise, Map<Integer, 
			List<Integer>> distance_freq, int min_range, Map<ChargeLocation, Integer> max_charge_times, Map<SocialWeight, Float> social_weights, float env_consciousness, 
			float enjoy_innovation, int income_category, int income, int age, int location, boolean homecharge, float recharge_moment, FuelType current_fuel_type, int current_vehicle_age, 
			int vehicle_in_possesion_age, int yearly_km_category, int yearly_km, List<Float> car_costs_range, boolean buys_new)
	{
		super(id, need_weights, ambition, uncertainty_tolerance, expertise, distance_freq, min_range, max_charge_times, 
				social_weights, env_consciousness, enjoy_innovation, income_category, income, age, location, homecharge, recharge_moment, current_fuel_type, current_vehicle_age, 
				vehicle_in_possesion_age, yearly_km_category, yearly_km, car_costs_range, buys_new, true);
	}	

	/* 
	 * ************************************************ NEEDS SATISFACTION FUNCTIONS ***********************************************
	 * 
	*/	
	
	public float computeYearlyCosts(String model, boolean totalLoss)
	{
		VehicleConcept vehicle_concept = this.getVehicleConcepts().get(model);
		FuelTypeConcept fueltype_concept =  this.getFuelTypeConcepts().get(vehicle_concept.getFuelType());
		
		// ::::FUEL::::
		float fuel_costs = fueltype_concept.getKmCosts() * d_yearly_km;
		
		//::::MAINTENANCE::::
		float failure_costs = fueltype_concept.getFailureCosts() * d_yearly_km;
		
		//::::BIJTELLING::::
		float price = vehicle_concept.getPrice();
		float taxes = ((float)Tax.getBijtellingLevel((int)vehicle_concept.getEmissions())/100) * (Tax.getIncomeTaxCategory(d_income_category)/100)  * price;
				
		float total_costs = fuel_costs + failure_costs + taxes;
		return total_costs;
	}
	
	/* 
	 * ***************************************************** NEW CAR FUNCTIONS ******************************************************
	 * 
	*/	
	
	public void decideToBuyNewVehicle()
	{				
		if (d_vehicle.batteryBroken())
			d_vehicle.buyNewBattery();
		
		// Lease Agent swaps car every 4 years
		if ((int)RepastEssentials.GetTickCount() - d_vehicle.getPurchaseTick() < (4 * 52) && d_vehicle.getKmCount() < 140000)
			return;
		
		// Else, select model to buy
		String model = getBestModel();
		
		// If no vehicle is available in agent's price class, do nothing yet
		if (model == "")
		{
			optimise();
			return;
		}

		// Trade own car for new model
		buyNewVehicle(model);
	}
	
	public String getBestModel()
	{
		// Check whether models have been updated
		Map<String, VehicleConcept> updatedVehicles = new HashMap<String, VehicleConcept>();
		for (String model : this.getVehicleConcepts().keySet()) 
		{
			Vehicle latestVersion = Marketplace.getLatestVersion(model);
			if (!this.getVehicleConcepts().containsKey(latestVersion.getModel()))
				updatedVehicles.put(latestVersion.getModel(), new VehicleConcept(latestVersion));
		}
		for (String model : updatedVehicles.keySet())
			getVehicleConcepts().put(model, updatedVehicles.get(model));
		
		Map<String, Float> model_with_yearly_costs = new HashMap<String, Float>();
		float cheapest_yearly_costs = -1;
		
		// For all known vehicle models
		for (String model : this.getVehicleConcepts().keySet())
		{	
			// Exclude unavailable (old) models
			if (!modelWithinPriceRange(model))
				continue;
			
			float yearly_costs = computeYearlyCosts(model, true);
			model_with_yearly_costs.put(model, yearly_costs);
			if (yearly_costs < cheapest_yearly_costs || cheapest_yearly_costs == -1)
				cheapest_yearly_costs = yearly_costs;
		}
		
		if (model_with_yearly_costs.isEmpty())
			return "";
		
		String best_model = "";
		float best_evaluation = 0;
		
		// Select best model available to the agent
		for (String model: model_with_yearly_costs.keySet())
		{			
			// Determine costs satisfaction
			float costs_satisfaction = Math.max(0, 1 - ((model_with_yearly_costs.get(model) - cheapest_yearly_costs) / cheapest_yearly_costs));
			
			// Determine overall satisfaction and uncertainty of this model.
			float satisfaction = computePerceivedNeedSatisfaction(model, costs_satisfaction);
			float certainty = (1 - computePerceivedNeedUncertainty(model));
			float evaluation = satisfaction * d_ambition + certainty * (1 - d_uncertainty_tolerance);
						
			// Check whether this model has highest rating.
			if (evaluation > best_evaluation || best_model == "")
			{
				best_model = model;
				best_evaluation = evaluation;
			}
		}		
		// Return model with highest rating.
		return best_model;
	}
	
	public boolean modelWithinPriceRange(String model)
	{
		return Marketplace.getNewPrice(model) >= 0;
	}
	
	public void buyNewVehicle(String model)
	{
		float currentMaintenanceCosts = d_vehicle.batteryBroken() ? d_vehicle.getFailureCosts() + d_vehicle.getNewBatteryCosts() : d_vehicle.getFailureCosts();		
		Marketplace.sellCar(d_vehicle,  d_vehicle.getValue() - currentMaintenanceCosts);
		d_vehicle = Marketplace.leaseNewCar(model);
		d_memoryown.changedOwnVehicle(d_vehicle);
	}
	
}
