/**
 * 
 */
package electric_vehicles.Agents;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import repast.simphony.essentials.RepastEssentials;
import electric_vehicles.Enums.ChargeLocation;
import electric_vehicles.Enums.FuelType;
import electric_vehicles.Enums.NeedWeight;
import electric_vehicles.Enums.SocialWeight;
import electric_vehicles.Memory.FuelTypeConcept;
import electric_vehicles.Memory.VehicleConcept;
import electric_vehicles.VehicleInfrastructure.Marketplace;
import electric_vehicles.VehicleInfrastructure.Tax;
import electric_vehicles.Vehicles.Vehicle;

/**
 * @author Ayla
 *
 */
public class AgentNew extends Agent {
			
	public AgentNew(int id, Map<NeedWeight,Float> need_weights, float ambition, float uncertainty_tolerance, float expertise, Map<Integer, 
			List<Integer>> distance_freq, int min_range, Map<ChargeLocation, Integer> max_charge_times, Map<SocialWeight, Float> social_weights, float env_consciousness, 
			float enjoy_innovation, int income_category, int income, int age, int location, boolean homecharge, float recharge_moment, FuelType current_fuel_type, 
			int current_vehicle_age, int vehicle_in_possesion_age, int yearly_km_category, int yearly_km, List<Float> car_costs_range, boolean buys_new)
	{
		super(id, need_weights, ambition, uncertainty_tolerance, expertise, distance_freq, min_range, max_charge_times, 
				social_weights, env_consciousness, enjoy_innovation, income_category, income, age, location, homecharge, recharge_moment, 
				current_fuel_type, current_vehicle_age, vehicle_in_possesion_age, yearly_km_category, yearly_km, car_costs_range, buys_new, false);
		
		d_money = (float) (d_income * (0.5 * d_money_prop) * (RepastEssentials.GetTickCount()+1 - d_vehicle.getPurchaseTick())/52);
		
	}
		
	/* 
	 * ************************************************ NEEDS SATISFACTION FUNCTIONS ***********************************************
	 * 
	*/	
	
	public float computeYearlyCosts(String model, boolean totalLoss)
	{
		VehicleConcept vehicle_concept = this.getVehicleConcepts().get(model);
		FuelTypeConcept fueltype_concept =  this.getFuelTypeConcepts().get(vehicle_concept.getFuelType());
		
		// ::::DEPRECIATION::::
		float price = !totalLoss && model == d_vehicle.getModel() ? d_vehicle.getValue() : vehicle_concept.getPrice();
		float depreciation = (float) ((price * (1 - Math.pow(0.85, 4)))/4);
				
		// ::::FUEL::::
		float fuel_costs = fueltype_concept.getKmCosts() * d_yearly_km;
				
		//::::MAINTENANCE::::
		float failure_costs = fueltype_concept.getFailureCosts() * d_yearly_km;
				
		//::::TAX::::
		float taxes = Tax.getTaxLevel((int)vehicle_concept.getEmissions()) * 12;
				
		float total_costs = depreciation + fuel_costs + failure_costs + taxes;
		
		return total_costs;
	}
	
	/* 
	 * ***************************************************** NEW CAR FUNCTIONS ******************************************************
	 * 
	*/	
	
	public void decideToBuyNewVehicle()
	{		
		// Increase money buffer
		d_money += d_income * d_money_prop * 1/52;
				
		// Determine this week's repair costs (possibly including costs of fixing an EV/PHEV's battery)
		float currentMaintenanceCosts = d_vehicle.batteryBroken() ? d_memoryown.getLastStepFailureCosts() + d_vehicle.getNewBatteryCosts() : d_memoryown.getLastStepFailureCosts();
		boolean vehicle_total_loss = d_vehicle.getValue()+100 > currentMaintenanceCosts ? false : true;
		
		// If agent is overall still satisfied and vehicle is worth more than last repair costs, pay any repair costs and continue
		if (d_memoryown.satisfiedWithCurrentVehicle() && !vehicle_total_loss)
		{
			// If battery is broken and agent has money to fix it, do so.
			if (d_vehicle.batteryBroken())// && d_money > d_vehicle.getNewBatteryCosts())
				buyNewBattery();
			if (d_vehicle.getFirstPurchaseTick() + 2 * 52 > RepastEssentials.GetTickCount()) 
				d_money -= d_memoryown.getLastStepFailureCosts();
			return;
		}
			
		// Else, select model to buy
		String model = getBestModel(vehicle_total_loss);
		
		// If agent's favorite model is too expensive/same as current model, pay any repair costs continue
		if (model == "")
		{
			// If battery is broken and agent has money to fix it, do so.
			if (d_vehicle.batteryBroken())// && d_money > d_vehicle.getNewBatteryCosts())
				buyNewBattery();
			if (d_vehicle.getFirstPurchaseTick() + 2 * 52 > RepastEssentials.GetTickCount()) 
				d_money -= d_memoryown.getLastStepFailureCosts();
			return;
		}

		// Trade own car for new model
		buyNewVehicle(model, vehicle_total_loss);
	}
	
	public String getBestModel(boolean vehicle_total_loss)
	{
		// Check whether models have been updated
		Map<String, VehicleConcept> updatedVehicles = new HashMap<String, VehicleConcept>();
		for (String model : this.getVehicleConcepts().keySet()) 
		{
			Vehicle latestVersion = Marketplace.getLatestVersion(model);
			if (!this.getVehicleConcepts().containsKey(latestVersion.getModel()))
				updatedVehicles.put(latestVersion.getModel(), new VehicleConcept(latestVersion));
		}
		for (String model : updatedVehicles.keySet())
			getVehicleConcepts().put(model, updatedVehicles.get(model));
		
		Map<String, Float> model_with_yearly_costs = new HashMap<String, Float>();
		float cheapest_yearly_costs = -1;
		
		// For all known vehicle models
		for (String model : this.getVehicleConcepts().keySet())
		{			
			// Model is not in right price segment: Exclude model. (unavailable models are -1, so are always excluded)
			if (!modelWithinPriceRange(model))
				continue;
			
			float yearly_costs = computeYearlyCosts(model, vehicle_total_loss);
			model_with_yearly_costs.put(model, yearly_costs);
			if (yearly_costs < cheapest_yearly_costs || cheapest_yearly_costs == -1)
				cheapest_yearly_costs = yearly_costs;
		}
		
		String best_model = "";
		float best_evaluation = 0;
		
		// Select best model available to the agent
		for (String model: model_with_yearly_costs.keySet())
		{
			// Determine costs satisfaction
			float costs_satisfaction = Math.max(0, 1 - ((model_with_yearly_costs.get(model) - cheapest_yearly_costs) / cheapest_yearly_costs));
			
			// Determine overall satisfaction and uncertainty of this model.
			float satisfaction = computePerceivedNeedSatisfaction(model, costs_satisfaction);
			float certainty = (1 - computePerceivedNeedUncertainty(model));
			float evaluation = satisfaction * d_ambition + certainty * (1 - d_uncertainty_tolerance);
						
			// Check whether this model has highest rating.
			if (evaluation > best_evaluation || best_model == "")
			{
				best_model = model;
				best_evaluation = evaluation;
			}
		}
		
		// If agents doesn't know any vehicle's in its price class. Don't buy anything.
		if (best_model == "")
			return best_model;
		
		// Don't buy same model again, unless current vehicle is beyond repair.
		if (best_model.equals(d_vehicle.getModel()) && !vehicle_total_loss)
			best_model = "";
		
		// If agent does not have enough money yet, then don't buy vehicle now unless current vehicle is beyond repair.
		else if (Marketplace.getNewPrice(best_model) - Math.max(100, d_vehicle.getValue() - d_memoryown.getLastStepFailureCosts()) > d_money && !vehicle_total_loss)
			best_model = "";
		
		// Return model with highest rating.
		return best_model;
	}
	
	public boolean modelWithinPriceRange(String model)
	{
		return (Marketplace.getNewPrice(model) >= d_car_costs_range.get(0) && Marketplace.getNewPrice(model) <= d_car_costs_range.get(1));
	}
	
	public void buyNewVehicle(String model, boolean vehicle_total_loss)
	{
		float currentMaintenanceCosts = d_vehicle.batteryBroken() ? d_memoryown.getLastStepFailureCosts() + d_vehicle.getNewBatteryCosts() : d_memoryown.getLastStepFailureCosts();
		
		d_money += Math.max(100, d_vehicle.getValue() - currentMaintenanceCosts);		// Obtain money for current vehicle's value, minus any repair costs, with a minimum of 100 (scrap value)
		Marketplace.sellCar(d_vehicle, d_vehicle.getValue() - currentMaintenanceCosts);	// Sell current car 
		d_vehicle = Marketplace.buyNewCar(model);										// Buy new car and pay for it
		d_money -= d_vehicle.getValue();
		d_memoryown.changedOwnVehicle(d_vehicle);										// Reset all aspects in memory related to previous vehicle
	}

	public void buyNewBattery()
	{
		d_vehicle.buyNewBattery();
		if (d_vehicle.getFirstPurchaseTick() + 8 * 52 > RepastEssentials.GetTickCount() || d_vehicle.getKmCount() > 100000)
			d_money -= d_vehicle.getNewBatteryCosts();
	}
}
