/**
 * 
 */
package electric_vehicles.Agents;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import repast.simphony.essentials.RepastEssentials;
import electric_vehicles.Enums.ChargeLocation;
import electric_vehicles.Enums.FuelType;
import electric_vehicles.Enums.NeedWeight;
import electric_vehicles.Enums.SocialWeight;
import electric_vehicles.Memory.FuelTypeConcept;
import electric_vehicles.Memory.VehicleConcept;
import electric_vehicles.VehicleInfrastructure.Marketplace;
import electric_vehicles.VehicleInfrastructure.Tax;
import electric_vehicles.Vehicles.Battery;
import electric_vehicles.Vehicles.Vehicle;

/**
 * @author Ayla
 *
 */
public class AgentOccasion extends Agent {
	
	private float d_income_weight;
	
	public AgentOccasion(int id, Map<NeedWeight,Float> need_weights, float ambition, float uncertainty_tolerance, float expertise, Map<Integer, 
			List<Integer>> distance_freq, int min_range, Map<ChargeLocation, Integer> max_charge_times, Map<SocialWeight, Float> social_weights, float env_consciousness, 
			float enjoy_innovation, int income_category, int income, int age, int location, boolean homecharge, float recharge_moment, FuelType current_fuel_type, 
			int current_vehicle_age, int vehicle_in_possesion_age, int yearly_km_category, int yearly_km, List<Float> car_costs_range, boolean buys_new)
	{		
		super(id, need_weights, ambition, uncertainty_tolerance, expertise, distance_freq, min_range, max_charge_times, 
				social_weights, env_consciousness, enjoy_innovation, income_category, income, age, location, homecharge, recharge_moment, current_fuel_type, 
				current_vehicle_age, vehicle_in_possesion_age, yearly_km_category, yearly_km, car_costs_range, buys_new, false);

		d_income_weight = (-(float)1/18) * d_income_category + ((float)29/36);	
		d_money = (float) (d_income * (0.5 * d_money_prop) * (RepastEssentials.GetTickCount()+1 - d_vehicle.getPurchaseTick())/52);				
	}	
	
	
	/* 
	 * ************************************************ NEEDS SATISFACTION FUNCTIONS ***********************************************
	 * 
	*/	
	
	public float computeYearlyCosts(String model, boolean totalLoss)
	{
		VehicleConcept vehicle_concept = this.getVehicleConcepts().get(model);
		FuelTypeConcept fueltype_concept =  this.getFuelTypeConcepts().get(vehicle_concept.getFuelType());
		
		
		// ::::DEPRECIATION::::
		float price = 0;
		price = !totalLoss && model == d_vehicle.getModel() ? d_vehicle.getValue() : (float) (Math.pow(0.85, vehicle_concept.getPrice() * 8));
		float depreciation = (float) ((price * (1 - Math.pow(0.85, 4)))/4);
				
		// ::::FUEL::::
		float fuel_costs = fueltype_concept.getKmCosts() * d_yearly_km;
				
		//::::MAINTENANCE::::
		float failure_costs = fueltype_concept.getFailureCosts() * d_yearly_km;
		if (vehicle_concept.getFuelType() == FuelType.ELECTRIC)
			failure_costs += ((vehicle_concept.getRange() * Battery.getFuelEconomyFactor()) * Battery.getKWhCosts()/((Battery.getBatteryLifeSlow() + Battery.getBatteryLifeFast())/2));
		
		//::::TAX::::
		float taxes = Tax.getTaxLevel((int)vehicle_concept.getEmissions()) * 12;
		
		float total_costs = depreciation + fuel_costs + failure_costs + taxes;
		
		return total_costs;
	}
	
	/* 
	 * ***************************************************** NEW CAR FUNCTIONS ******************************************************
	 * 
	*/	
	
	public void decideToBuyNewVehicle()
	{		
		// Increase money buffer
		d_money += d_income * d_money_prop * 1/52;
				
		// Determine this week's repair costs (possibly including costs of fixing an EV/PHEV's battery)
		float totalMaintenanceCosts = d_vehicle.batteryBroken() ? d_memoryown.getLastStepFailureCosts() + d_vehicle.getNewBatteryCosts() : d_memoryown.getLastStepFailureCosts();
		boolean vehicle_total_loss = d_vehicle.getValue()+100 > totalMaintenanceCosts ? false : true;
				
		// If agent is overall still satisfied and vehicle is worth more than last repair costs, pay any repair costs and continue
		if (d_memoryown.satisfiedWithCurrentVehicle() && !vehicle_total_loss)
		{
			// If battery is broken and agent has money to fix it, do so.
			if (d_vehicle.batteryBroken())// && d_money > d_vehicle.getNewBatteryCosts())
				buyNewBattery();
			if (d_vehicle.getFirstPurchaseTick() + 2 * 52 > RepastEssentials.GetTickCount()) 
				d_money -= d_memoryown.getLastStepFailureCosts();
			return;
		}
		// Else, select model to buy
		String model = getBestModel(vehicle_total_loss);
				
		// If agent's favorite model is too expensive/same as current model or no model is available, pay any repair costs and continue
		if (model == "")
		{
			// If battery is broken, fix it (making this dependent on agent's money results in problems. agents in debt may never buy a new battery).
			if (d_vehicle.batteryBroken())// && d_money > d_vehicle.getNewBatteryCosts())
				buyNewBattery();
			if (d_vehicle.getFirstPurchaseTick() + 2 * 52 > RepastEssentials.GetTickCount()) 
				d_money -= d_memoryown.getLastStepFailureCosts();
			return;
		}

		// Check which occasion is best
		List<Vehicle> available_cars = Marketplace.getOccasionsFromModel(model);
		int index = getBestOccasionOption(available_cars, model);
		
		// Trade own car for new model
		buyOccasionVehicle(model, index, vehicle_total_loss);
	}
	
	public String getBestModel(boolean vehicle_total_loss)
	{
		Map<String, Float> model_with_yearly_costs = new HashMap<String, Float>();
		float cheapest_yearly_costs = -1;
		
		// Select models within price range and calculate those's yearly costs
		for (String model : this.getVehicleConcepts().keySet())
		{
			
			// Model does not exist in right price segment in occasion market, check if it occurs in new market. 
			if (!modelWithinPriceRange(model))
					continue;	

			float yearly_costs = computeYearlyCosts(model, vehicle_total_loss);
			model_with_yearly_costs.put(model, yearly_costs);
			if (yearly_costs < cheapest_yearly_costs || cheapest_yearly_costs == -1)
				cheapest_yearly_costs = yearly_costs;
		}
		
		String best_model = "";
		float best_evaluation = 0;
		
		// Select best model available to the agent
		for (String model: model_with_yearly_costs.keySet())
		{
			// Determine costs satisfaction
			float costs_satisfaction = Math.max(0, 1 - ((model_with_yearly_costs.get(model) - cheapest_yearly_costs) / cheapest_yearly_costs));
			
			// Determine overall satisfaction and uncertainty of this model.
			float satisfaction = computePerceivedNeedSatisfaction(model, costs_satisfaction);
			float certainty = (1 - computePerceivedNeedUncertainty(model));
			float evaluation = satisfaction * d_ambition + certainty * (1 - d_uncertainty_tolerance);
						
			// Check whether this model has highest rating.
			if (evaluation > best_evaluation || best_model == "")
			{
				best_model = model;
				best_evaluation = evaluation;
			}
		}
		
		// If agents doesn't know any vehicle's in its price class. Don't buy anything.
		if (best_model == "")
			return best_model;
			
		// Don't buy same model again, unless current vehicle is beyond repair.
		if (best_model.equals(d_vehicle.getModel()) && !vehicle_total_loss)
			best_model = "";

		// If agent does not have enough money for occasion yet; don't buy it unless current vehicle is beyond repair.
		else if (Marketplace.cheapestVehicleInOccasionMarketInPriceRange(best_model, d_car_costs_range) - Math.max(100, d_vehicle.getValue() - d_memoryown.getLastStepFailureCosts()) > d_money && !vehicle_total_loss)
			best_model = "";
		
		// Return model with highest rating.
		return best_model;
	}
	
	public boolean modelWithinPriceRange(String model)
	{
		return Marketplace.occasionMarketHasModelInPriceRange(model, d_car_costs_range);
	}
	
	public int getBestOccasionOption(List<Vehicle> vehicles, String model)
	{		
		// If only one vehicle is available, buy this one
		if (vehicles.size() == 1)
			return 0;
		
		float total_value = 0;
		float total_km = 0;
		
		for(Vehicle vehicle : vehicles)
		{
			total_value += vehicle.getValue();
			total_km += vehicle.getKmCount();
		}
		
		// Provide each vehicle with a rating, according to its normalized price and km count.
		List<Float> vehicle_rating = new ArrayList<Float>();
		for(Vehicle vehicle : vehicles)
		{
			// If vehicle is more expensive than budget, don't bother
			if (vehicle.getValue() < (d_car_costs_range.get(0) - 2500) || vehicle.getValue() > d_car_costs_range.get(1))
				vehicle_rating.add(0.0f);
			// Else determine vehicle rating based on its price and km count
			else
			{
				float rating = d_income_weight * ( 1 - (vehicle.getValue() / total_value) ) + (1 - d_income_weight) * ( 1 - (vehicle.getKmCount() / total_km));
				vehicle_rating.add(rating);
			}
		}
			
		// Select vehicle with highest rating
		int best_option = -1;
		float best_rating = 0.0f;
		for (int i = 0; i < vehicle_rating.size(); ++i)
		{
			if (vehicle_rating.get(i) > best_rating)
			{
				best_option = i;
				best_rating = vehicle_rating.get(i);
			}
		}
		
		return best_option;
	}
	
	/* Trade own car for new model (temp new_vehicle is used so index of occasion market is not messed up)*/
	public void buyOccasionVehicle(String model, int index, boolean vehicle_total_loss)
	{	
		float currentMaintenanceCosts = d_vehicle.batteryBroken() ? d_memoryown.getLastStepFailureCosts() + d_vehicle.getNewBatteryCosts() : d_memoryown.getLastStepFailureCosts();
		
		Vehicle new_vehicle = Marketplace.buyOccasionCar(model, index);
		
		d_money += Math.max(100, d_vehicle.getValue() - currentMaintenanceCosts);		// Obtain money for current vehicle's value, minus any repair costs, with a minimum of 100 (scrap value)
		Marketplace.sellCar(d_vehicle,  d_vehicle.getValue() - currentMaintenanceCosts);// Sell current car 
		d_vehicle = new_vehicle;														// Buy new car and pay for it
		d_money -= d_vehicle.getValue();
		d_memoryown.changedOwnVehicle(d_vehicle);										// Reset all aspects in memory related to previous vehicle
	}
	
	
	public void buyNewBattery()
	{
		d_vehicle.buyNewBattery();
		if (d_vehicle.getFirstPurchaseTick() + 8 * 52 > RepastEssentials.GetTickCount() || d_vehicle.getKmCount() > 100000)
			d_money -= d_vehicle.getNewBatteryCosts();
	}

}
