/**
 * 
 */
package electric_vehicles.Media;

import java.util.HashMap;
import java.util.Map;

import repast.simphony.engine.schedule.ScheduledMethod;
import repast.simphony.essentials.RepastEssentials;
import repast.simphony.random.RandomHelper;
import electric_vehicles.Enums.ChargeOption;
import electric_vehicles.Enums.FuelType;
import electric_vehicles.VehicleInfrastructure.Marketplace;
import electric_vehicles.VehicleInfrastructure.Mechanic;
import electric_vehicles.VehicleInfrastructure.RechargeStation;
import electric_vehicles.Vehicles.Vehicle;

/**
 * @author Ayla
 *
 */
public class MediaWriter {

	protected Map<FuelType, Float> d_truth_deviation; 
	protected Map<FuelType, Float> d_write_probability; 
	
	public MediaWriter()
	{
		d_truth_deviation = new HashMap<FuelType, Float>();
		d_write_probability = new HashMap<FuelType, Float>();
	}
	
	@ScheduledMethod(start = 1 , interval = 1, priority = 4)
	public void createNewMediaMessages()
	{
		for (FuelType type : FuelType.values())
		{
			if (RandomHelper.nextDoubleFromTo(0, 1) > 0.5)
				continue;
			
			float truth_deviation = d_truth_deviation.get(type) == 0 ? 0 : (float) RandomHelper.nextDoubleFromTo(0, d_truth_deviation.get(type));
			
			// Obtain general variable values	
			float km_costs = RechargeStation.getKmCosts(type, ChargeOption.SLOW) * (1 - truth_deviation);
			float failure_costs = Mechanic.getAverageFailureCostsPerKm(type) * (1 - truth_deviation);
			float charge_time = RechargeStation.getChargeTime(type, ChargeOption.SLOW) * (1 - truth_deviation);
			float charge_time_fast = RechargeStation.getChargeTime(type, ChargeOption.FAST) * (1 - truth_deviation);
	
			// Add media message to archive
			Media.publishMessage(new MediaMessage(type, km_costs, failure_costs, charge_time, charge_time_fast));
		}
		
		for (Vehicle vehicle : Marketplace.getAllModels())
		{
			float time_since_introduction = (int)RepastEssentials.GetTickCount() - vehicle.getIntroductionTick();
			float write_probability = 1 - ((float)time_since_introduction / 104);
			if (RandomHelper.nextDoubleFromTo(0, 1) > write_probability)
				continue;
			
			FuelType type = vehicle.getFuelType();
			
			float truth_deviation = d_truth_deviation.get(type) == 0 ? 0 : (float) RandomHelper.nextDoubleFromTo(0, d_truth_deviation.get(type));
			
			// Obtain vehicle's characteristics	
			String model = vehicle.getModel();
			float price = vehicle.getPrice() * (1 - truth_deviation);
			float range = vehicle.getRange() * (1 + truth_deviation);
			float emissions = vehicle.getEmissions() * (1 - truth_deviation);
						
			Media.publishMessage(new MediaMessage(type, model, price, range, emissions));
		}
	}
	
}
