/**
 * 
 */
package electric_vehicles.Memory;

import repast.simphony.random.RandomHelper;
import electric_vehicles.Enums.ChargeOption;
import electric_vehicles.Enums.FuelType;
import electric_vehicles.VehicleInfrastructure.Mechanic;
import electric_vehicles.VehicleInfrastructure.RechargeStation;
import electric_vehicles.Vehicles.Vehicle;

/**
 * @author Ayla
 *
 */
public class FuelTypeConcept {
	
	private int d_count;

	// Fuel properties
	private FuelType d_fuelType;
	
	private float d_kmCosts;	
	private float d_failureCosts;
	private float d_chargeTime;
	private float d_chargeTimeFast;
	
	private float d_kmCosts_variance;	
	private float d_failureCosts_variance;	
	private float d_chargeTime_variance;
	private float d_chargeTimeFast_variance;
	
	// Constructor when fuel types are initialized.
	public FuelTypeConcept(FuelType fuelType, float expertise)
	{
		d_count = 1;
		d_fuelType = fuelType;
			
		float deviation = 1 - expertise;
				
		float kmCosts = RechargeStation.getKmCosts(fuelType, ChargeOption.SLOW);
		d_kmCosts = (float) (kmCosts - RandomHelper.nextDoubleFromTo(-(deviation * kmCosts), deviation * kmCosts));
		
		float failureCosts = Mechanic.getAverageFailureCostsPerKm(fuelType);
		d_failureCosts = (float) (failureCosts - RandomHelper.nextDoubleFromTo(-(deviation * failureCosts), deviation * failureCosts));
		
		float chargeTime = RechargeStation.getChargeTime(fuelType, ChargeOption.SLOW);
		d_chargeTime = (float) (chargeTime - RandomHelper.nextDoubleFromTo(-(deviation * chargeTime), deviation * chargeTime));
		
		float chargeTimeFast = RechargeStation.getChargeTime(fuelType, ChargeOption.FAST);
		d_chargeTimeFast = (float) (chargeTimeFast - RandomHelper.nextDoubleFromTo(-(deviation * chargeTimeFast), deviation * chargeTimeFast));
				
		d_kmCosts_variance = 0;
		d_failureCosts_variance = 0;
		d_chargeTime_variance = 0;
		d_chargeTimeFast_variance = 0;
	}
	
	// Constructor when parsing a media message or fuel type concept of another agent.
	public FuelTypeConcept(FuelType fuelType, float kmCosts, float failureCosts, float chargeTime, float chargeTimeFast)
	{
		d_count = 1;
		d_fuelType = fuelType;
		
		d_kmCosts = kmCosts;
		d_failureCosts = failureCosts;
		d_chargeTime = chargeTime;
		d_chargeTimeFast = chargeTimeFast;
		
		d_kmCosts_variance = 0;
		d_failureCosts_variance = 0;
		d_chargeTime_variance = 0;
		d_chargeTimeFast_variance = 0;
	}
	
	public FuelTypeConcept copy()
	{
		return new FuelTypeConcept(d_fuelType, d_kmCosts, d_failureCosts, d_chargeTime, d_chargeTimeFast);
	}
	
	public int getCount()					{ return d_count; }
	public FuelType getFuelType()			{ return d_fuelType; }
	
	public float getKmCosts()				{ return d_kmCosts; }
	public float getFailureCosts()			{ return d_failureCosts; }
	public float getChargeTime()			{ return d_chargeTime; }
	public float getChargeTimeFast()		{ return d_chargeTimeFast; }	

	/* Update estimated mean and variance of various fuel stats,
	 */
	public void updateStats(FuelTypeConcept fuel, float weight)
	{
		if (!fuel.getFuelType().equals(d_fuelType))
			System.out.println("ERROR: OTHER. Trying to update wrong fuel type concept. This FTC: '" + d_fuelType + "' received: '" + fuel.getFuelType() + "'");
		
		d_count++;
		
		float old_value_weight = 1 - weight/10;
		float new_value_weight = weight/10;
		
		float delta = fuel.getKmCosts() - d_kmCosts;
		d_kmCosts = d_kmCosts * old_value_weight + fuel.getKmCosts() * new_value_weight;
		d_kmCosts_variance = d_kmCosts_variance * old_value_weight + (delta * (fuel.getKmCosts() - d_kmCosts)) * new_value_weight;
		
		delta = fuel.getFailureCosts() - d_failureCosts;
		d_failureCosts = d_failureCosts * old_value_weight + fuel.getFailureCosts() * new_value_weight;
		d_failureCosts_variance = d_failureCosts_variance * old_value_weight + (delta * (fuel.getFailureCosts() - d_failureCosts)) * new_value_weight;
		
		delta = fuel.getChargeTime() - d_chargeTime;
		d_chargeTime = d_chargeTime * old_value_weight + fuel.getChargeTime() * new_value_weight;
		d_chargeTime_variance = d_chargeTime_variance * old_value_weight + (delta * (fuel.getChargeTime() - d_chargeTime)) * new_value_weight;
		
		delta = fuel.getChargeTimeFast() - d_chargeTimeFast;
		d_chargeTimeFast = d_chargeTimeFast * old_value_weight + fuel.getChargeTimeFast() * new_value_weight;
		d_chargeTimeFast_variance = d_chargeTimeFast_variance * old_value_weight + (delta * (fuel.getChargeTimeFast() - d_chargeTimeFast)) * new_value_weight;
	}
	
	public void updateStatsOwnVehicle(Vehicle vehicle)
	{
		if (!vehicle.getFuelType().equals(d_fuelType))
			System.out.println("ERROR: OWN. Trying to update wrong fuel type concept. This FTC: '" + d_fuelType + "' received: '" + vehicle.getFuelType() + "'");
		
		d_count++;
		
		float delta = vehicle.getCurrentKmCosts() - d_kmCosts;
		d_kmCosts = d_kmCosts * 0.9f + vehicle.getCurrentKmCosts() * 0.1f;
		d_kmCosts_variance = d_kmCosts_variance * 0.9f + (delta * (vehicle.getCurrentKmCosts() - d_kmCosts)) * 0.1f;
		
		delta = vehicle.getAverageFailureCosts() - d_failureCosts;
		d_failureCosts = d_failureCosts * 0.9f + vehicle.getAverageFailureCosts() * 0.1f;
		d_failureCosts_variance = d_failureCosts_variance * 0.9f + (delta * (vehicle.getAverageFailureCosts() - d_failureCosts)) * 0.1f;
		
		float chargeTime = RechargeStation.getChargeTime(vehicle.getFuelType(), ChargeOption.SLOW);
		delta = chargeTime - d_chargeTime;
		d_chargeTime = d_chargeTime * 0.9f + chargeTime * 0.1f;
		d_chargeTime_variance = d_chargeTime_variance * 0.9f + (delta * (chargeTime - d_chargeTime)) * 0.1f;
		
		float chargeTimeFast = RechargeStation.getChargeTime(vehicle.getFuelType(), ChargeOption.FAST);
		delta = chargeTimeFast - d_chargeTimeFast;
		d_chargeTimeFast = d_chargeTimeFast * 0.9f + chargeTimeFast * 0.1f;
		d_chargeTimeFast_variance = d_chargeTimeFast_variance * 0.9f + (delta * (chargeTimeFast - d_chargeTimeFast)) * 0.1f;
		
	}
	
	public float getKmCostsUncertainty()
	{
		return (float) Math.min(1, (d_count <= 1 || d_kmCosts == 0) ? 0 : Math.sqrt(d_kmCosts_variance) / d_kmCosts);
	}
	
	public float getFailureCostsUncertainty()
	{
		return (float) Math.min(1, (d_count <= 1 || d_failureCosts == 0) ? 0 : Math.sqrt(d_failureCosts_variance) / d_failureCosts);
	}
	
	public float getChargeTimeUncertainty()
	{
		if (d_count <= 1)
			return 0;
		
		float chargeTimeSlow_uncertainty = (float) Math.min(1, d_chargeTime == 0 ? 0 : Math.sqrt(d_chargeTime_variance) / d_chargeTime);
		float chargeTimeFast_uncertainty = (float) Math.min(1, d_chargeTimeFast == 0 ? 0 : Math.sqrt(d_chargeTimeFast_variance) / d_chargeTimeFast);
		
		return (chargeTimeSlow_uncertainty + chargeTimeFast_uncertainty)/2;
	}
	
}
