/**
 * 
 */
package electric_vehicles.Memory;


import electric_vehicles.Enums.FuelType;
import electric_vehicles.VehicleInfrastructure.Tax;
import electric_vehicles.Vehicles.Vehicle;

/**
 * @author Ayla
 *
 */
public class VehicleConcept {
	
	private int d_count;

	// Vehicle properties
	private FuelType d_fuelType;
	private String d_model;	
	
	private float d_range;
	private float d_price;
	private float d_emissions;
	private float d_taxes;
	private float d_bijtelling;
	
	private float d_range_variance;
	private float d_price_variance;
	private float d_emissions_variance;
	private float d_taxes_variance;
	private float d_bijtelling_variance;

	
	public VehicleConcept(FuelType fuelType, String model, float price, float range, float emissions)
	{
		d_count = 1;
		
		d_fuelType = fuelType;
		d_model = model;
		
		d_range = range;
		d_price = price;
		d_emissions = emissions;
		d_taxes = Tax.getTaxLevel((int)emissions);
		d_bijtelling = Tax.getBijtellingLevel((int)emissions);
		
		d_range_variance = 0;
		d_price_variance = 0;
		d_emissions_variance = 0;
		d_taxes_variance = 0;
		d_bijtelling_variance = 0;
	}
	
	public VehicleConcept(Vehicle vehicle)
	{
		d_count = 0;
		
		d_fuelType = vehicle.getFuelType();
		d_model = vehicle.getModel();
		
		d_range = vehicle.getRange();
		d_price = vehicle.getPrice();
		d_emissions = vehicle.getEmissions();
		d_taxes = Tax.getTaxLevel(vehicle.getEmissions());
		d_bijtelling = Tax.getBijtellingLevel(vehicle.getEmissions());
		
		d_range_variance = 0;
		d_price_variance = 0;
		d_emissions_variance = 0;
		d_taxes_variance = 0;
		d_bijtelling_variance = 0;
	}
	
	public VehicleConcept copy()
	{
		return new VehicleConcept(d_fuelType, d_model, d_price, d_range, d_emissions);
	}
	
	// Getters
	public int getCount()				{ return d_count; }
	public FuelType getFuelType()		{ return d_fuelType; }
	public String getModel()			{ return d_model; }
	public float getRange()				{ return d_range; }
	public float getPrice()				{ return d_price; }
	public float getEmissions()			{ return d_emissions; }

	
	/* Update estimated mean and variance of various vehicle stats,
	 */
	public void updateStats(VehicleConcept vehicle, float weight)
	{
		if (!vehicle.getModel().equals(d_model))
			System.out.println("ERROR: OTHER. Trying to update wrong vehicle concept. This VC: '" + d_model + "' received: '" + vehicle.getModel() + "'");
		
		d_count++;
		
		float old_value_weight = 1 - weight/10;
		float new_value_weight = weight/10;
		
		float delta = vehicle.getRange() - d_range;
		d_range = d_range * old_value_weight + vehicle.getRange() * new_value_weight;
		d_range_variance = d_range_variance * old_value_weight + (delta * (vehicle.getRange() - d_range)) * new_value_weight;
		
		delta = vehicle.getPrice() - d_price;
		d_price = d_price * old_value_weight + vehicle.getPrice() * new_value_weight;
		d_price_variance = d_price_variance * old_value_weight + (delta * (vehicle.getPrice() - d_price)) * new_value_weight;
		
		delta = vehicle.getEmissions() - d_emissions;
		d_emissions = d_emissions * old_value_weight + vehicle.getEmissions() * new_value_weight;
		d_emissions_variance = d_emissions_variance * old_value_weight + (delta * (vehicle.getEmissions() - d_emissions)) * new_value_weight;
		
		float taxes = Tax.getTaxLevel((int)vehicle.getEmissions());
		delta =  taxes - d_taxes;
		d_taxes = d_taxes * old_value_weight + taxes * new_value_weight;
		d_taxes_variance = d_taxes_variance * old_value_weight + (delta * (taxes - d_taxes)) * new_value_weight;
		
		taxes = Tax.getBijtellingLevel((int)vehicle.getEmissions());
		delta =  taxes - d_bijtelling;
		d_bijtelling = d_bijtelling * old_value_weight + taxes * new_value_weight;
		d_bijtelling_variance = d_bijtelling_variance * old_value_weight + (delta * (taxes - d_bijtelling)) * new_value_weight;
	}
	
	public void updateStatsOwnVehicle(Vehicle vehicle)
	{
		if (!vehicle.getModel().equals(d_model))
			System.out.println("ERROR: OWN. Trying to update wrong vehicle concept. This VC: '" + d_model + "' received: '" + vehicle.getModel() + "'");
		
		d_count++;
		
		float delta = vehicle.getRange() - d_range;
		d_range = d_range * 0.9f + vehicle.getRange() * 0.1f;
		d_range_variance = d_range_variance * 0.9f + (delta * (vehicle.getRange() - d_range)) * 0.1f;
		
		d_price = vehicle.getPrice();
		d_emissions = vehicle.getEmissions();
		
		float taxes = Tax.getTaxLevel((int)vehicle.getEmissions());
		delta = taxes - d_taxes;
		d_taxes = d_taxes * 0.9f + taxes* 0.1f;
		d_taxes_variance = d_taxes_variance * 0.9f + (delta * (taxes - d_taxes)) * 0.1f;
		
		taxes = Tax.getBijtellingLevel((int)vehicle.getEmissions());
		delta = taxes - d_bijtelling;
		d_bijtelling = d_bijtelling * 0.9f + taxes * 0.1f;
		d_bijtelling_variance = d_bijtelling_variance * 0.9f + (delta * (taxes - d_bijtelling)) * 0.1f;
	}
		
	public float getRangeUncertainty()
	{
		return (float) Math.min(1, (d_count <= 1 || d_range == 0) ? 0 : Math.sqrt(d_range_variance) / d_range);
	}
	
	public float getPriceUncertainty()
	{
		return (float) Math.min(1, (d_count <= 1 || d_price == 0) ? 0 : Math.sqrt(d_price_variance) / d_price);
	}
	
	public float getTaxUncertainty(boolean lease)
	{
		if (lease)
			return (float) Math.min(1, (d_count <= 1 || d_bijtelling == 0) ? 0 : Math.sqrt(d_bijtelling_variance) / d_bijtelling);
		return (float) Math.min(1, (d_count <= 1 || d_taxes == 0) ? 0 : Math.sqrt(d_taxes_variance) / d_taxes);
	}
	
	public float getPreferenceUncertainty()
	{	
		return (float) Math.min(1, (d_count <= 1 || d_emissions == 0) ? 0 : Math.sqrt(d_emissions_variance) / d_emissions);
	}
}
