/**
 * 
 */
package electric_vehicles.Scenarios;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import repast.simphony.engine.schedule.ScheduledMethod;
import repast.simphony.essentials.RepastEssentials;
import electric_vehicles.UserPanel;
import electric_vehicles.Enums.ChargeLocation;
import electric_vehicles.Enums.ChargeOption;
import electric_vehicles.Enums.FuelType;
import electric_vehicles.Enums.InstructionType;
import electric_vehicles.Scenarios.Instructions.BatteryCostsInstruction;
import electric_vehicles.Scenarios.Instructions.BatteryLifeInstruction;
import electric_vehicles.Scenarios.Instructions.BijtellingInstruction;
import electric_vehicles.Scenarios.Instructions.EmissionUpdateInstruction;
import electric_vehicles.Scenarios.Instructions.FuelCostsInstruction;
import electric_vehicles.Scenarios.Instructions.FuelTimeInstruction;
import electric_vehicles.Scenarios.Instructions.InfrastructureInstruction;
import electric_vehicles.Scenarios.Instructions.Instruction;
import electric_vehicles.Scenarios.Instructions.MaintenanceCostsInstruction;
import electric_vehicles.Scenarios.Instructions.MoneyInstruction;
import electric_vehicles.Scenarios.Instructions.TaxInstruction;
import electric_vehicles.Scenarios.Instructions.VehicleInstruction;
import electric_vehicles.VehicleInfrastructure.Marketplace;
import electric_vehicles.VehicleInfrastructure.TaxLevel;
import electric_vehicles.Vehicles.Battery;
import electric_vehicles.Vehicles.ElectricVehicle;
import electric_vehicles.Vehicles.GasolineVehicle;
import electric_vehicles.Vehicles.HybridVehicle;
import electric_vehicles.Vehicles.Vehicle;

/**
 * @author Ayla
 *
 */
public class Scenarios {
	
	// Initialization variables
	public static String d_name;
	private static int d_stop = 5200;
	private static float d_emissions_adjustment;
	private String d_scenario_file_path;
	
	// Instruction archives
	private Map<Integer, List<Instruction> > d_future_instructions;
	private Map<Integer, List<Instruction> > d_instructions;
	
	public static void setEmissionsUpdate(float yearly_adjustment)	{ d_emissions_adjustment = yearly_adjustment;	}
	public static int getDuration()									{ return d_stop; }
	
	/* 
	 * ******************************************************* PARSE SCENARIO FILE ******************************************************
	 * 
	*/
	
	public Scenarios()
	{
		d_future_instructions = new HashMap<Integer, List<Instruction> >();
		d_instructions = new HashMap<Integer, List<Instruction> >();
		d_emissions_adjustment = 1;
		
		switch(UserPanel.scenario)
		{
			case "Verification": 		System.out.println("Running Verification Scenario");
										d_scenario_file_path = "scenario/verification/verification.csv";
										break;
			case "Basic": 				System.out.println("Running Basic Scenario");
										d_scenario_file_path = "scenario/basic.csv";
										break;
			case "All": 				System.out.println("Running All Scenario");
										d_scenario_file_path = "scenario/all.csv";
										break;
			case "A1 Bijtelling":		System.out.println("Running Government: Bijtelling Scenario");
										d_scenario_file_path = "scenario/A_Government/1_Bijtelling.csv";
										break;
			case "A1 Bijtelling BEV":	System.out.println("Running Government: Bijtelling BEV Scenario");
										d_scenario_file_path = "scenario/A_Government/1_Bijtelling_BEV.csv";
										break;
			case "A1 Bijtelling Gov":	System.out.println("Running Government: Bijtelling BEV Government Scenario");
										d_scenario_file_path = "scenario/A_Government/1_Bijtelling_Gov.csv";
										break;
			case "A2 Gasoline price":	System.out.println("Running Government: Gasoline price Scenario");
										d_scenario_file_path = "scenario/A_Government/2_GasolinePrice.csv";
										break;
			case "A3 Fast charge price":System.out.println("Running Government: Fast charge price Scenario");
										d_scenario_file_path = "scenario/A_Government/3_FastChargePrice.csv";
										break;
			case "A23 Prices":			System.out.println("Running Government: Fast Charge & Gasoline price Scenario");
										d_scenario_file_path = "scenario/A_Government/2_3_Prices.csv";
										break;
			case "A4 BPM":				System.out.println("Running Government: BPM Scenario");
										d_scenario_file_path = "scenario/A_Government/4_BPM.csv";
										break;	
			case "A4 BPM BEV":			System.out.println("Running Government: BPM BEV Scenario");
										d_scenario_file_path = "scenario/A_Government/4_BPM_BEV.csv";
										break;	
			case "B1 Fast charge time": System.out.println("Running Technology: Fast charge time Scenario");
										d_scenario_file_path = "scenario/B_Technology/1_FastChargeTime.csv";
										break;
			case "B2 Battery costs":	System.out.println("Running Technology: Battery costs Scenario");
										d_scenario_file_path = "scenario/B_Technology/2_BatteryCosts.csv";
										break;
			case "B3 Fast charge prob":	System.out.println("Running Technology: Fast charge probability Scenario");
										d_scenario_file_path = "scenario/B_Technology/3_FastChargeProb.csv";
										break;
			case "C1 Fast Charge":		System.out.println("Running Combination: Fast charge Scenario");
										d_scenario_file_path = "scenario/C_Combi/1_FastCharge.csv";
										break;
			case "C1 Fast Charge ProbAndCosts":System.out.println("Running Combination: Fast charge Costs Scenario");
										d_scenario_file_path = "scenario/C_Combi/1_FastCharge_Costs.csv";
										break;			
			case "C1 Fast Charge Time":	System.out.println("Running Combination: Fast charge Time Scenario");
										d_scenario_file_path = "scenario/C_Combi/1_FastCharge_Time.csv";
										break;
			case "C2 CombiPHEV":		System.out.println("Running Combination: All for PHEV Scenario");
										d_scenario_file_path = "scenario/C_Combi/2_CombiPHEV.csv";
										break;
			case "C2 CombiPHEV Ext":	System.out.println("Running Combination: All for PHEV Extended Scenario");
										d_scenario_file_path = "scenario/C_Combi/2_CombiPHEV_Ext.csv";
										break;	
			case "C3 CombiBEV":		System.out.println("Running Combination: All for BEV Scenario");
										d_scenario_file_path = "scenario/C_Combi/3_CombiBEV.csv";
										break;
            default:					System.out.println("No scenario selected. Running Basic Scenario");
										d_scenario_file_path = "scenario/basic.csv";
										break;
		}
		initializeScenario();
	}
	
	public void initializeScenario()
	{
		File data_file = new File(d_scenario_file_path);
		BufferedReader reader = null;	
		try
		{
			reader = new BufferedReader(new FileReader(data_file));
			String row = null;
			
			List<String> labels = new LinkedList<String>();
			InstructionType category = InstructionType.GENERAL;
			
			// Parse instructions
			while ((row = reader.readLine()) != null)
			{
				// Split data from this instruction
				List<String> data_elements = Arrays.asList(row.split(";"));
				
				// Skip empty lines
				if (data_elements.isEmpty())
					continue;
				
				// New instruction category reached? 
				if (data_elements.get(0).equals("#"))
				{
					category = InstructionType.fromString((data_elements.get(1)));
					// Parse data labels
					labels = data_elements;
					continue;
				}
				
				// Start of document?
				if (category == InstructionType.GENERAL)
				{
					parseGeneralInstruction(labels, data_elements);
					continue;
				}
				
				// End of document?
				if (category == InstructionType.END)
					return;
				
				// Add tick to instruction archive
				int tick = Integer.parseInt(data_elements.get(labels.indexOf("Start")));
				if (!d_future_instructions.containsKey(tick))
					d_future_instructions.put(tick, new LinkedList<Instruction>());
				
				// Parse instruction
				switch(category)
				{
				case BATTERY_COSTS:		parseBatteryCostsInstruction(labels, data_elements, tick);
								 		break;
				case BATTERY_LIFE:		parseBatteryLifeInstruction(labels, data_elements, tick);
		 								break;
				case BIJTELLING:		parseBijtellingInstruction(labels, data_elements, tick);
										break;
				case EMISSIONS_UPGRADE:	parseEmissionsUpgradeInstruction(labels, data_elements, tick);
										break;
				case FUEL_COSTS:		parseFuelCostsInstruction(labels, data_elements, tick);
										break;
				case FUEL_TIME:			parseFuelTimeInstruction(labels, data_elements, tick);
										break;
				case INFRASTRUCTURE:	parseInfrastructureInstruction(labels, data_elements, tick);
										break;
				case MAINTENANCE:		parseMaintenanceInstruction(labels, data_elements, tick);
										break;
				case MONEY:				parseMoneyInstruction(labels, data_elements, tick);
										break;
				case TAX:				parseTaxInstruction(labels, data_elements, tick);
										break;
				case VEHICLE:			parseVehicleInstruction(labels, data_elements, tick);
										break;
				default:				System.out.println("ERROR: Unrecognized instruction category");
										break;
				}
			}
		} 
		catch (FileNotFoundException e) 
		{
		    e.printStackTrace();
		}
		catch (IOException e) 
		{
		    e.printStackTrace();
		} 
		finally 
		{
		    try 
		    {
		        if (reader != null) 
		            reader.close();
		    } 
		    catch (IOException e) 
		    {}
		}
	}
	
	/* 
	 * ******************************************************* PARSE INSTRUCTIONS ******************************************************
	 * 
	*/
	
	public void parseBatteryCostsInstruction(List<String> labels, List<String> data, int start)
	{
		int stop = Integer.parseInt(data.get(labels.indexOf("Stop")));
		int adjustment = Integer.parseInt(data.get(labels.indexOf("Adjustment")));
		
		d_future_instructions.get(start).add(new BatteryCostsInstruction(start, stop, adjustment));
	}
	
	public void parseBatteryLifeInstruction(List<String> labels, List<String> data, int start)
	{
		int stop = Integer.parseInt(data.get(labels.indexOf("Stop")));
		int adjustment_min = Integer.parseInt(data.get(labels.indexOf("Adjustment min")));
		int adjustment_max = Integer.parseInt(data.get(labels.indexOf("Adjustment max")));
		
		d_future_instructions.get(start).add(new BatteryLifeInstruction(start, stop, adjustment_min, adjustment_max));
	}
	
	public void parseBijtellingInstruction(List<String> labels, List<String> data, int start)
	{
		List<TaxLevel> bijtelling_levels = new LinkedList<TaxLevel>();
		int min_emissions = 0;
		int max_bijtelling = 0;
		
		for (int i = 1; i <= 8; ++i)
		{
			// Stop once empty category is reached
			if (labels.indexOf("CAT"+i) > data.size() - 1)
				break;
			
			// Obtain information on a bijtelling's level
			int bijtelling =  Integer.parseInt(data.get(labels.indexOf("BIJT"+i)));
			int max_emissions = Integer.parseInt(data.get(labels.indexOf("CAT"+i))) + 1; 
			
			// Add bijtelling level with specified emission boundaries
			bijtelling_levels.add(new TaxLevel(bijtelling, min_emissions, max_emissions));
			min_emissions = max_emissions;
			
			// Store highest known bijtelling level
			max_bijtelling = bijtelling > max_bijtelling ? bijtelling : max_bijtelling;
		}
		int max_emmissions = min_emissions;
				
		d_future_instructions.get(start).add(new BijtellingInstruction(bijtelling_levels, max_bijtelling, max_emmissions));		
	}
	
	public void parseEmissionsUpgradeInstruction(List<String> labels, List<String> data, int start)
	{
		int stop = Integer.parseInt(data.get(labels.indexOf("Stop")));
		float yearly_adjustment = Float.parseFloat(data.get(labels.indexOf("Yearly adjustment")));
		d_future_instructions.get(start).add(new EmissionUpdateInstruction(yearly_adjustment));
		
		if (!d_future_instructions.containsKey(start+stop))
			d_future_instructions.put(start+stop, new LinkedList<Instruction>());
		d_future_instructions.get(start+stop).add(new EmissionUpdateInstruction(1));
	}
	
	public void parseFuelCostsInstruction(List<String> labels, List<String> data, int start)
	{
		int stop = Integer.parseInt(data.get(labels.indexOf("Stop")));
		ChargeOption type = ChargeOption.fromString(data.get(labels.indexOf("Type")));
		float adjustment = Float.parseFloat(data.get(labels.indexOf("Adjustment")));
		
		d_future_instructions.get(start).add(new FuelCostsInstruction(start, stop, type, adjustment));
	}
	
	public void parseFuelTimeInstruction(List<String> labels, List<String> data, int start)
	{
		int stop = Integer.parseInt(data.get(labels.indexOf("Stop")));
		ChargeOption type = ChargeOption.fromString(data.get(labels.indexOf("Type")));
		int adjustment = Integer.parseInt(data.get(labels.indexOf("Adjustment")));
		
		d_future_instructions.get(start).add(new FuelTimeInstruction(start, stop, type, adjustment));
	}
	
	public void parseGeneralInstruction(List<String> labels, List<String> data)
	{
		d_name = data.get(labels.indexOf("Name"));
		d_stop = Integer.parseInt(data.get(labels.indexOf("Stop")));
	}
	
	public void parseInfrastructureInstruction(List<String> labels, List<String> data, int start)
	{
		int stop = Integer.parseInt(data.get(labels.indexOf("Stop")));
		ChargeLocation charge_location = ChargeLocation.fromString((data.get(labels.indexOf("Location"))));
		float adjustment = Float.parseFloat(data.get(labels.indexOf("Adjustment")));
		
		d_future_instructions.get(start).add(new InfrastructureInstruction(start, stop, charge_location, adjustment));
	}
	
	public void parseMaintenanceInstruction(List<String> labels, List<String> data, int start)
	{
		int stop = Integer.parseInt(data.get(labels.indexOf("Stop")));
		FuelType type = FuelType.fromString(data.get(labels.indexOf("Type")));
		float adjustment = Float.parseFloat(data.get(labels.indexOf("Adjustment")));
		
		d_future_instructions.get(start).add(new MaintenanceCostsInstruction(start, stop, type, adjustment));
	}
	
	public void parseMoneyInstruction(List<String> labels, List<String> data, int start)
	{
		int stop = Integer.parseInt(data.get(labels.indexOf("Stop")));
		float adjustment = Float.parseFloat(data.get(labels.indexOf("Adjustment")));
		
		d_future_instructions.get(start).add(new MoneyInstruction(start, stop, adjustment));
	}
	
	public void parseTaxInstruction(List<String> labels, List<String> data, int start)
	{
		List<TaxLevel> tax_levels = new LinkedList<TaxLevel>();
		int min_emissions = 0;
		int max_tax = 0;
		
		for (int i = 1; i <= 8; ++i)
		{		
			// Stop once empty category is reached
			if (labels.indexOf("CAT"+i) > data.size() - 1)
				break;
			
			// Obtain information on a tax's level
			int tax =  Integer.parseInt(data.get(labels.indexOf("TAX"+i)));
			int max_emissions = Integer.parseInt(data.get(labels.indexOf("CAT"+i))) + 1; 
			
			// Add tax level with specified emission boundaries
			tax_levels.add(new TaxLevel(tax, min_emissions, max_emissions));
			min_emissions = max_emissions;
			
			// Store highest known tax level
			max_tax = tax > max_tax ? tax : max_tax;
		}

		d_future_instructions.get(start).add(new TaxInstruction(tax_levels, max_tax));		
	}
	
	/* Parses which vehicles must be introduced when 
	 * */
	public void parseVehicleInstruction(List<String> labels, List<String> data, int start)
	{		
		FuelType fuelType = FuelType.fromString(data.get(labels.indexOf("Fuel type")));
		String model = data.get(labels.indexOf("Model"));
		float price = Float.parseFloat(data.get(labels.indexOf("Price")));
		int emissions = Integer.parseInt(data.get(labels.indexOf("Emissions")));
		int range = fuelType == FuelType.ELECTRIC ? 0 : Integer.parseInt(data.get(labels.indexOf("Range")));
		int capacity = fuelType == FuelType.GASOLINE ? 0 : Integer.parseInt(data.get(labels.indexOf("Battery capacity")));

		d_future_instructions.get(start).add(new VehicleInstruction(fuelType, model, price, emissions, range, capacity));
	}
	
	/* 
	 * ******************************************************* EXECUTE INSTRUCTIONS ******************************************************
	 * 
	*/
	
	@ScheduledMethod(start = 1 , interval = 1, priority = 5)
	public void executeInstructions()
	{
		int current_tick = (int)RepastEssentials.GetTickCount();
				
		if (current_tick == d_stop)
			RepastEssentials.EndSimulationRun();

		// Activate new instructions
		if (d_future_instructions.containsKey(current_tick))
		{
			List<Instruction> instructions = d_future_instructions.remove(current_tick);
			for (Instruction instruction : instructions)
			{	
				int stop_tick = instruction.activate();
				
				// If this instruction spreads over time, store it
				for (int i = current_tick + 1; i < stop_tick; ++i)
				{
					if (!d_instructions.containsKey(i))
						d_instructions.put(i, new LinkedList<Instruction>());
					d_instructions.get(i).add(instruction);
				}
			}
		}
	
		// Execute past instructions which spread over time
		if (d_instructions.containsKey(current_tick))
		{
			List<Instruction> instructions = d_instructions.remove(current_tick);
			for (Instruction instruction : instructions)
				instruction.execute();
		}
		
		// Update existing models in marketplace
		//if (d_emissions_adjustment != 1)
		//{
			List<Vehicle> all_vehicles = Marketplace.getAllModels();
			List<String> outdated_models = new LinkedList<String>();
			List<Vehicle> new_models = new LinkedList<Vehicle>();
			
			for (Vehicle vehicle : all_vehicles)
			{
				if ((current_tick - vehicle.getIntroductionTick()) != 52)
					continue;

				switch (vehicle.getFuelType())
				{
				case GASOLINE:	new_models.add(new GasolineVehicle(FuelType.GASOLINE, vehicle.getModelSeries(), current_tick, vehicle.getPrice(), vehicle.getRange(), (int)(vehicle.getEmissions()*d_emissions_adjustment)));
								break;
				case HYBRID:	new_models.add(new HybridVehicle(FuelType.HYBRID, vehicle.getModelSeries(), current_tick, vehicle.getPrice(), vehicle.getBatteryPrice(), vehicle.getRange(), (int)(vehicle.getEmissions()*d_emissions_adjustment), (int)(vehicle.getBatteryPrice() / Battery.getKWhCosts())));
								break;
				case ELECTRIC:	new_models.add(new ElectricVehicle(FuelType.ELECTRIC, vehicle.getModelSeries(), current_tick, vehicle.getPrice(), vehicle.getBatteryPrice(), vehicle.getEmissions(), (int)(vehicle.getBatteryPrice() / Battery.getKWhCosts())));
								break;
				default:		break;	
				}
				outdated_models.add(vehicle.getModel());
			}
			
			for (String model : outdated_models)
				Marketplace.removeModel(model);
			
			for (Vehicle vehicle : new_models)
				Marketplace.introduceCar(vehicle);
		//}

	}
	

}
