package electric_vehicles;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import electric_vehicles.Agents.Agent;
import electric_vehicles.Enums.FuelType;
import electric_vehicles.Enums.InformationStrategy;
import electric_vehicles.Enums.NeedWeight;
import electric_vehicles.Scenarios.Scenarios;
import electric_vehicles.VehicleInfrastructure.Marketplace;
import electric_vehicles.VehicleInfrastructure.RechargeStation;
import electric_vehicles.Vehicles.Vehicle;
import repast.simphony.context.Context;
import repast.simphony.engine.schedule.ScheduledMethod;
import repast.simphony.essentials.RepastEssentials;
import repast.simphony.util.ContextUtils;
import repast.simphony.util.collections.IndexedIterable;

public class Statistics {
	
	String d_file_path;
	
	// Market
	File d_new_sales_file, d_occ_sales_file;
	File d_gasOcc_file, d_EVOcc_file, d_PHEVOcc_file;
	File d_total_scrap_file, d_scrap_age_file; 
	// Ownership
	File d_debt_file, d_age_file, d_duration_file, d_emissions_file, d_fastCharged;
	// Diffusion
	File d_newGas_file, d_newEV_file, d_newPHEV_file;
	Map<FuelType, File> d_total, d_totalPerNewAgent, d_totalPerLeaseAgent, d_totalPerOccAgent;
	// Perception
	Map<FuelType, File> d_PercSatisfaction;
	Map<FuelType, File> d_PercSatisCost, d_PercSatisFunc, d_PercSatisSoc, d_PercSatisPref;
	Map<FuelType, File> d_PercUncertainty;
	Map<FuelType, File> d_PercUncerCost, d_PercUncerFunc, d_PercUncerSoc, d_PercUncerPref;
	// Experience
	File d_IsSatisfiedNewAgent;
	Map<FuelType, File> d_ExpSatisfaction, d_IsSatisfied;
	Map<FuelType, File> d_ExpSatisCost, d_ExpSatisFunc, d_ExpSatisSoc, d_ExpSatisPref;
	Map<FuelType, File> d_ExpUncertainty;
	Map<FuelType, File> d_ExpUncerCost, d_ExpUncerFunc, d_ExpUncerSoc, d_ExpUncerPref;
	// Availability
	File d_knowsEVleaser, d_knowsPHEVleaser, d_knowsEVtotal, d_knowsPHEVtotal;
	// Strategies
	Map<FuelType, File> d_imitation, d_inquiring, d_repitition, d_optimising;
	// Extra
	File d_EVOccMedianValue_file, d_PHEVOccMedianValue_file, d_gasOccMedianValue_file;	
	File d_EVOccMeanValue_file, d_PHEVOccMeanValue_file, d_gasOccMeanValue_file;
	File d_newMeanPrice_file, d_leaseMeanPrice_file;
	File d_EVOccMeanAge_file, d_PHEVOccMeanAge_file, d_gasOccMeanAge_file;
	File d_allEVOccValue_file, d_allPHEVOccValue_file, d_allGasOccValue_file;
	File d_allEVOccAge_file, d_allPHEVOccAge_file, d_allGasOccAge_file;
	// Paper
	Map<FuelType, File> d_leassee, d_income;
	Map<FuelType, File> d_yearlyKm, d_hasHomeCharge, d_hasWorkCharge, d_ambition, d_uncTolerance;
	Map<FuelType, File> d_wFinancial, d_wFunctional, d_wSocial, d_wEnvironmental;
	
	public Statistics() 
	{
		d_file_path = "results/" + Scenarios.d_name + "/";
		// Market
		d_new_sales_file = createFile(d_file_path + "Market/newSales.csv");
		d_occ_sales_file = createFile(d_file_path + "Market/occSales.csv");
		d_gasOcc_file = createFile(d_file_path + "Market/gasOcc.csv");
		d_EVOcc_file = createFile(d_file_path + "Market/EVOcc.csv");
		d_PHEVOcc_file = createFile(d_file_path + "Market/PHEVOcc.csv");
		d_total_scrap_file = createFile(d_file_path + "Market/totalScrap.csv");
		d_scrap_age_file = createFile(d_file_path + "Market/scrapAge.csv");
		// Ownership
		d_debt_file = createFile(d_file_path + "Owner/debt.csv");
		d_emissions_file = createFile(d_file_path + "Owner/emissions.csv");
		d_age_file = createFile(d_file_path + "Owner/age.csv");
		d_duration_file = createFile(d_file_path + "Owner/duration.csv");
		d_fastCharged = createFile(d_file_path + "Owner/fastcharged.csv");
		// Diffusion
		d_newGas_file = createFile(d_file_path + "Diffusion/newGas.csv");
		d_newEV_file = createFile(d_file_path + "Diffusion/newEV.csv");
		d_newPHEV_file = createFile(d_file_path + "Diffusion/newPHEV.csv");
		d_total = new HashMap<FuelType, File>();
		d_totalPerNewAgent = new HashMap<FuelType, File>();
		d_totalPerLeaseAgent = new HashMap<FuelType, File>();
		d_totalPerOccAgent = new HashMap<FuelType, File>();
		// Perception
		d_PercSatisfaction = new HashMap<FuelType, File>();
		d_PercUncertainty = new HashMap<FuelType, File>();
		d_PercSatisFunc = new HashMap<FuelType, File>();
		d_PercSatisCost = new HashMap<FuelType, File>();
		d_PercSatisSoc = new HashMap<FuelType, File>();
		d_PercSatisPref = new HashMap<FuelType, File>();
		d_PercUncerFunc = new HashMap<FuelType, File>();
		d_PercUncerCost = new HashMap<FuelType, File>();
		d_PercUncerSoc = new HashMap<FuelType, File>();
		d_PercUncerPref = new HashMap<FuelType, File>();
		// Experience
		d_IsSatisfiedNewAgent = createFile(d_file_path + "Experience/IsSatisfiedNewAgent.csv");
		d_IsSatisfied = new HashMap<FuelType, File>();
		d_ExpSatisfaction = new HashMap<FuelType, File>();
		d_ExpSatisCost = new HashMap<FuelType, File>();
		d_ExpSatisFunc = new HashMap<FuelType, File>();
		d_ExpSatisSoc = new HashMap<FuelType, File>();
		d_ExpSatisPref = new HashMap<FuelType, File>();
		d_ExpUncertainty = new HashMap<FuelType, File>();
		d_ExpUncerCost = new HashMap<FuelType, File>();
		d_ExpUncerFunc = new HashMap<FuelType, File>();
		d_ExpUncerSoc = new HashMap<FuelType, File>();
		d_ExpUncerPref = new HashMap<FuelType, File>();
		// Availability
		d_knowsEVleaser = createFile(d_file_path + "Availability/knowsEVLeaser.csv");
		d_knowsPHEVleaser = createFile(d_file_path + "Availability/knowsPHEVLeaser.csv");
		d_knowsEVtotal = createFile(d_file_path + "Availability/knowsEVtotal.csv");
		d_knowsPHEVtotal = createFile(d_file_path + "Availability/knowsPHEVtotal.csv");
		// Strategies
		d_imitation = new HashMap<FuelType, File>(); 
		d_inquiring = new HashMap<FuelType, File>(); 
		d_repitition = new HashMap<FuelType, File>(); 
		d_optimising = new HashMap<FuelType, File>(); 
		// Extra
		d_EVOccMeanValue_file = createFile(d_file_path + "Extra/MeanValueEVOcc.csv");
		d_PHEVOccMeanValue_file = createFile(d_file_path + "Extra/MeanValuePHEVOcc.csv");
		d_gasOccMeanValue_file = createFile(d_file_path + "Extra/MeanValueGasOcc.csv");
		d_EVOccMedianValue_file = createFile(d_file_path + "Extra/MedianValueEVOcc.csv");
		d_PHEVOccMedianValue_file = createFile(d_file_path + "Extra/MedianValuePHEVOcc.csv");
		d_gasOccMedianValue_file = createFile(d_file_path + "Extra/MedianValueGasOcc.csv");
		d_newMeanPrice_file = createFile(d_file_path + "Extra/MeanPriceNew.csv");
		d_leaseMeanPrice_file = createFile(d_file_path + "Extra/MeanPriceLease.csv");
		d_EVOccMeanAge_file = createFile(d_file_path + "Extra/MeanAgeEVOcc.csv");
		d_PHEVOccMeanAge_file = createFile(d_file_path + "Extra/MeanAgePHEVOcc.csv");
		d_gasOccMeanAge_file = createFile(d_file_path + "Extra/MeanAgeGasOcc.csv");
		d_allEVOccValue_file = createFile(d_file_path + "Extra/allValueEVOcc.csv");
		d_allPHEVOccValue_file = createFile(d_file_path + "Extra/allValuePHEVOcc.csv");
		d_allGasOccValue_file = createFile(d_file_path + "Extra/allValueGasOcc.csv");
		d_allEVOccAge_file = createFile(d_file_path + "Extra/allAgeEVOcc.csv");
		d_allPHEVOccAge_file = createFile(d_file_path + "Extra/allAgePHEVOcc.csv");
		d_allGasOccAge_file = createFile(d_file_path + "Extra/allAgeGasOcc.csv");
		// Paper 
		d_hasHomeCharge = new HashMap<FuelType, File>();
		d_hasWorkCharge = new HashMap<FuelType, File>();
		d_yearlyKm = new HashMap<FuelType, File>();
		d_ambition = new HashMap<FuelType, File>();
		d_uncTolerance = new HashMap<FuelType, File>();
		d_leassee = new HashMap<FuelType, File>();
		d_income = new HashMap<FuelType, File>();
		d_wFinancial = new HashMap<FuelType, File>(); 
		d_wFunctional = new HashMap<FuelType, File>(); 
		d_wSocial = new HashMap<FuelType, File>(); 
		d_wEnvironmental = new HashMap<FuelType, File>();
	}
	
	@ScheduledMethod(start = 1, interval = 1, priority = 0)
	public void updateStatistics()
	{		
		// Market
		if (RepastEssentials.GetTickCount() == 1 || RepastEssentials.GetTickCount() % 52 == 0)
		{
			writeToFile(d_newMeanPrice_file, "New mean price", Float.toString(Marketplace.averageNewPrice()).replace('.', ','));
			writeToFile(d_leaseMeanPrice_file, "Lease mean price", Float.toString(Marketplace.averageLeasePrice()).replace('.', ','));
			writeToFile(d_new_sales_file, "New Sales", Integer.toString(Marketplace.getNewSales()));
			writeToFile(d_occ_sales_file, "Occasion Sales", Integer.toString(Marketplace.getOccasionSales()));
			writeToFile(d_total_scrap_file, "Total scrap", Integer.toString(Marketplace.getTotalLoss()));
			writeToFile(d_scrap_age_file, "Scrap age", Float.toString(Marketplace.averageScrapAge()).replace('.', ','));
			Marketplace.resetYear();
		}
		writeToFile(d_gasOcc_file, "Gasoline Occasions", Integer.toString(Marketplace.GasolineOccasionsCount()));
		writeToFile(d_EVOcc_file, "EV Occasions", Integer.toString(Marketplace.EVsOccasionsCount()));
		writeToFile(d_PHEVOcc_file, "PHEV Occasions", Integer.toString(Marketplace.HybridOccasionsCount()));

		
		// Ownership
		Map<String, Double> ownerStats = updateOwnerStats();
		writeToFile(d_debt_file, "Debt", Double.toString(ownerStats.get("Debt")).replace('.', ','));
		writeToFile(d_emissions_file, "Emissions", Double.toString(ownerStats.get("Emissions")).replace('.', ','));
		writeToFile(d_age_file, "Vehicle age", Double.toString(ownerStats.get("Age")).replace('.', ','));
		writeToFile(d_duration_file, "Owner duration", Double.toString(ownerStats.get("Duration")).replace('.', ','));
		writeToFile(d_fastCharged, "Fast charged energy", Integer.toString(RechargeStation.getFastChargedEnergy()).replace('.', ','));
		RechargeStation.resetFastChargedEnergy();
		
		// Diffusion
		if (RepastEssentials.GetTickCount() ==1 || RepastEssentials.GetTickCount() % 13 == 0)
		{
			writeToFile(d_newGas_file, "New Gasoline", Float.toString(Marketplace.getProportionGasSales()).replace('.', ','));
			writeToFile(d_newEV_file, "New EV", Float.toString(Marketplace.getProportionEVSales()).replace('.', ','));
			writeToFile(d_newPHEV_file, "New PHEV", Float.toString(Marketplace.getProportionPHEVSales()).replace('.', ','));
			Marketplace.resetQuarter();
		}
		writeToFileFTs(d_total, "Diffusion", "Total", ownerStats);
		writeToFileFTs(d_totalPerNewAgent, "Diffusion", "TotalPerNewAgent", ownerStats);
		writeToFileFTs(d_totalPerLeaseAgent, "Diffusion", "TotalPerLeaseAgent", ownerStats);
		writeToFileFTs(d_totalPerOccAgent, "Diffusion", "TotalPerOccAgent", ownerStats);
		
		// Perception
		writeToFileFTs(d_PercSatisfaction, "Perception", "PercSatisfaction", ownerStats);
		writeToFileFTs(d_PercSatisFunc, "Perception", "PercSatisFunc", ownerStats);
		writeToFileFTs(d_PercSatisCost, "Perception", "PercSatisCost", ownerStats);
		writeToFileFTs(d_PercSatisSoc, "Perception", "PercSatisSoc", ownerStats);
		writeToFileFTs(d_PercSatisPref, "Perception", "PercSatisPref", ownerStats);
		writeToFileFTs(d_PercUncertainty, "Perception", "PercUncertainty", ownerStats);
		writeToFileFTs(d_PercUncerFunc, "Perception", "PercUncerFunc", ownerStats);
		writeToFileFTs(d_PercUncerCost, "Perception", "PercUncerCost", ownerStats);
		writeToFileFTs(d_PercUncerSoc, "Perception", "PercUncerSoc", ownerStats);
		writeToFileFTs(d_PercUncerPref, "Perception", "PercUncerPref", ownerStats);

		// Experience
		writeToFileFTs(d_IsSatisfied, "Experience", "IsSatisfied", ownerStats);
		writeToFile(d_IsSatisfiedNewAgent, "Experience", Double.toString(ownerStats.get("IsSatisfiedNewAgent")).replace('.', ','));
		writeToFileFTs(d_ExpSatisfaction, "Experience", "ExpSatisfaction", ownerStats);
		writeToFileFTs(d_ExpSatisCost, "Experience", "ExpSatisCost", ownerStats);
		writeToFileFTs(d_ExpSatisFunc, "Experience", "ExpSatisFunc", ownerStats);
		writeToFileFTs(d_ExpSatisSoc, "Experience", "ExpSatisSoc", ownerStats);
		writeToFileFTs(d_ExpSatisPref, "Experience", "ExpSatisPref", ownerStats);
		writeToFileFTs(d_ExpUncertainty, "Experience", "ExpUncertainty", ownerStats);
		writeToFileFTs(d_ExpUncerCost, "Experience", "ExpUncerCost", ownerStats);
		writeToFileFTs(d_ExpUncerFunc, "Experience", "ExpUncerFunc", ownerStats);
		writeToFileFTs(d_ExpUncerSoc, "Experience", "ExpUncerSoc", ownerStats);
		writeToFileFTs(d_ExpUncerPref, "Experience", "ExpUncerPref", ownerStats);
		
		// Availability
		writeToFile(d_knowsEVleaser, "KnowsEVLease", Double.toString(ownerStats.get("knows"+FuelType.ELECTRIC+"lease")).replace('.', ','));
		writeToFile(d_knowsPHEVleaser, "KnowsPHEVLease", Double.toString(ownerStats.get("knows"+FuelType.HYBRID+"lease")).replace('.', ','));
		writeToFile(d_knowsEVtotal, "KnowsEVtotal", Double.toString(ownerStats.get("knows"+FuelType.ELECTRIC+"total")).replace('.', ','));
		writeToFile(d_knowsPHEVtotal, "KnowsPHEVtotal", Double.toString(ownerStats.get("knows"+FuelType.HYBRID+"total")).replace('.', ','));
	
		// Strategies
		writeToFileFTs(d_imitation, "Strategies", InformationStrategy.IMITATION.toString(), ownerStats);
		writeToFileFTs(d_inquiring, "Strategies", InformationStrategy.INQUIRING.toString(), ownerStats);
		writeToFileFTs(d_repitition, "Strategies", InformationStrategy.REPETITION.toString(), ownerStats);
		writeToFileFTs(d_optimising, "Strategies", InformationStrategy.OPTIMISING.toString(), ownerStats);
		
		// Extra
		writeToFile(d_EVOccMedianValue_file, "EV Occasion median value", Float.toString(Marketplace.getMedianElectricOccValue()).replace('.', ','));
		writeToFile(d_PHEVOccMedianValue_file, "PHEV Occasion median value", Float.toString(Marketplace.getMedianHybridOccValue()).replace('.', ','));
		writeToFile(d_gasOccMedianValue_file, "Gasoline Occasion median value", Float.toString(Marketplace.getMedianGasolineOccValue()).replace('.', ','));
		writeToFile(d_EVOccMeanValue_file, "EV Occasion mean value", Float.toString(Marketplace.getMeanElectricOccValue()).replace('.', ','));
		writeToFile(d_PHEVOccMeanValue_file, "PHEV Occasion mean value", Float.toString(Marketplace.getMeanHybridOccValue()).replace('.', ','));
		writeToFile(d_gasOccMeanValue_file, "Gasoline Occasion mean value", Float.toString(Marketplace.getMeanGasolineOccValue()).replace('.', ','));
		writeToFile(d_EVOccMeanAge_file, "EV Occasion mean age", Float.toString(Marketplace.getMeanElectricOccAge()).replace('.', ','));
		writeToFile(d_PHEVOccMeanAge_file, "PHEV Occasion mean age", Float.toString(Marketplace.getMeanHybridOccAge()).replace('.', ','));
		writeToFile(d_gasOccMeanAge_file, "Gasoline Occasion mean age", Float.toString(Marketplace.getMeanGasolineOccAge()).replace('.', ','));
		if (RepastEssentials.GetTickCount() ==1 ||(int)RepastEssentials.GetTickCount() == 3) {
			for (float value : Marketplace.getAllGasolineOccValues()) {
				writeToFile(d_allGasOccValue_file, "Gasoline Occasion value", Float.toString(value).replace('.', ','));
			}
			for (float age : Marketplace.getAllGasolineOccAges()) {
				writeToFile(d_allGasOccAge_file, "Gasoline Occasion age", Float.toString(age).replace('.', ','));
			}
		}
		if (RepastEssentials.GetTickCount() ==1 ||(int)RepastEssentials.GetTickCount() == 1090) {
			for (float value : Marketplace.getAllElectricOccValues()) {
				writeToFile(d_allEVOccValue_file, "Electric Occasion value", Float.toString(value).replace('.', ','));
			}
			for (float age : Marketplace.getAllElectricOccAges()) {
				writeToFile(d_allEVOccAge_file, "Electric Occasion age", Float.toString(age).replace('.', ','));
			}
			for (float value : Marketplace.getAllHybridOccValues()) {
				writeToFile(d_allPHEVOccValue_file, "Hybrid Occasion value", Float.toString(value).replace('.', ','));
			}
			for (float age : Marketplace.getAllHybridOccAges()) {
				writeToFile(d_allPHEVOccAge_file, "Hybrid Occasion age", Float.toString(age).replace('.', ','));
			}
		}
		
		// Paper
		writeToFileFTs(d_hasHomeCharge, "Paper", "hasHomeCharge", ownerStats);
		writeToFileFTs(d_hasWorkCharge, "Paper", "hasWorkCharge", ownerStats);
		writeToFileFTs(d_yearlyKm, "Paper", "YearlyKm", ownerStats);
		writeToFileFTs(d_ambition, "Paper", "Ambition", ownerStats);
		writeToFileFTs(d_uncTolerance, "Paper", "uncTolerance", ownerStats);
		writeToFileFTs(d_leassee, "Paper", "leassee", ownerStats);
		writeToFileFTs(d_income, "Paper", "income", ownerStats);
		writeToFileFTs(d_wFinancial, "Paper", "wFinancial", ownerStats);
		writeToFileFTs(d_wFunctional, "Paper", "wFunctional", ownerStats);
		writeToFileFTs(d_wSocial, "Paper", "wSocial", ownerStats);
		writeToFileFTs(d_wEnvironmental, "Paper", "wEnvironmental", ownerStats);		
	}
	

	
	public Map<String, Double> updateOwnerStats()
	{
		@SuppressWarnings("unchecked")
		Context<Agent> context = (Context<Agent>)ContextUtils.getContext(this);			
		IndexedIterable<Agent> agents = (IndexedIterable<Agent>)context.getObjects(Agent.class);	
		double debt = 0;
		double emissions = 0;
		double distance = 0;
		double vehicleAge = 0;
		double ownDuration = 0;
		Map<String, Double> ownerStats = new HashMap<String, Double>();
		Map<String, Integer> knownCount = new HashMap<String, Integer>();
		ownerStats.put("IsSatisfiedNewAgent", 0.0);
		for (FuelType ft : FuelType.values())
		{
			knownCount.put(ft.toString(), 0);
			ownerStats.put(ft + "Total", 0.0);
			ownerStats.put(ft + "TotalPerNewAgent", 0.0);
			ownerStats.put(ft + "TotalPerLeaseAgent", 0.0);
			ownerStats.put(ft + "TotalPerOccAgent", 0.0);

			ownerStats.put(ft + "IsSatisfied", 0.0);
			
			ownerStats.put(ft + "ExpSatisfaction", 0.0);
			ownerStats.put(ft + "ExpSatisCost", 0.0);
			ownerStats.put(ft + "ExpSatisFunc", 0.0);
			ownerStats.put(ft + "ExpSatisSoc", 0.0);
			ownerStats.put(ft + "ExpSatisPref", 0.0);
			
			ownerStats.put(ft + "ExpUncertainty", 0.0);
			ownerStats.put(ft + "ExpUncerCost", 0.0);
			ownerStats.put(ft + "ExpUncerFunc", 0.0);
			ownerStats.put(ft + "ExpUncerSoc", 0.0);
			ownerStats.put(ft + "ExpUncerPref", 0.0);
						
			ownerStats.put(ft + "PercSatisfaction", 0.0);
			ownerStats.put(ft + "PercSatisFunc", 0.0);
			ownerStats.put(ft + "PercSatisCost", 0.0);
			ownerStats.put(ft + "PercSatisSoc", 0.0);
			ownerStats.put(ft + "PercSatisPref", 0.0);
			
			ownerStats.put(ft + "PercUncertainty", 0.0);
			ownerStats.put(ft + "PercUncerFunc", 0.0);
			ownerStats.put(ft + "PercUncerCost", 0.0);
			ownerStats.put(ft + "PercUncerSoc", 0.0);
			ownerStats.put(ft + "PercUncerPref", 0.0);
			
			ownerStats.put(ft + InformationStrategy.IMITATION.toString(), 0.0);
			ownerStats.put(ft + InformationStrategy.INQUIRING.toString(), 0.0);
			ownerStats.put(ft + InformationStrategy.REPETITION.toString(), 0.0);
			ownerStats.put(ft + InformationStrategy.OPTIMISING.toString(), 0.0);
			
			ownerStats.put(ft + "hasHomeCharge", 0.0);
			ownerStats.put(ft + "hasWorkCharge", 0.0);
			ownerStats.put(ft + "YearlyKm", 0.0);
			ownerStats.put(ft + "Ambition", 0.0);
			ownerStats.put(ft + "uncTolerance", 0.0);
			ownerStats.put(ft + "leassee", 0.0);
			ownerStats.put(ft + "income", 0.0);
			ownerStats.put(ft + "wFinancial", 0.0);
			ownerStats.put(ft + "wFunctional", 0.0);
			ownerStats.put(ft + "wSocial", 0.0);
			ownerStats.put(ft + "wEnvironmental", 0.0);
		}
		ownerStats.put("knows"+FuelType.ELECTRIC+"lease", 0.0);
		ownerStats.put("knows"+FuelType.HYBRID+"lease", 0.0);
		int totalKnowsEV = 0;
		int totalKnowsPHEV = 0;
		int buysNew = 0;
		int leasers = 0;
		int occasioners = 0;
			
		// Access each agent
		for(Agent agent : agents)
		{
			boolean knowsPHEV = false;
			boolean knowsEV = false;
					
			// Determine easy-access characteristics
			debt += agent.inDebt();
			emissions += agent.getWeeklyEmissions();
			distance += agent.getWeeklyDistance();
			vehicleAge += agent.vehicleAge();
			ownDuration += agent.averageOwnershipTime();
			
			// Determine experiences of fuel types
			FuelType ft = agent.getFuelType();
			ownerStats.put(ft + "Total", ownerStats.get(ft + "Total") + 1);
			if (agent.isLeaser()) {
				ownerStats.put(ft + "TotalPerLeaseAgent", ownerStats.get(ft + "TotalPerLeaseAgent") + 1);
				leasers++;
			}
			else if (agent.buysNew()) {
				ownerStats.put(ft + "TotalPerNewAgent", ownerStats.get(ft + "TotalPerNewAgent") + 1);
				ownerStats.put("IsSatisfiedNewAgent", ownerStats.get("IsSatisfiedNewAgent") + agent.satisfiedWithCurrentVehicle());
				buysNew++;
			}
			else {
				ownerStats.put(ft + "TotalPerOccAgent", ownerStats.get(ft + "TotalPerOccAgent") + 1);
				occasioners++;
			}
			ownerStats.put(ft + "IsSatisfied", ownerStats.get(ft + "IsSatisfied") + agent.satisfiedWithCurrentVehicle());
			
			if (agent.hasHomeCharge())
				ownerStats.put(ft + "hasHomeCharge", ownerStats.get(ft + "hasHomeCharge") + 1);
			if (agent.hasWorkCharge())
				ownerStats.put(ft + "hasWorkCharge", ownerStats.get(ft + "hasWorkCharge") + 1);
			ownerStats.put(ft + "YearlyKm", ownerStats.get(ft + "YearlyKm") + agent.getKmPerYear());
			ownerStats.put(ft + "Ambition", ownerStats.get(ft + "Ambition") + agent.getAmbition());
			ownerStats.put(ft + "uncTolerance", ownerStats.get(ft + "uncTolerance") + agent.getUncertaintyTolerance());
			ownerStats.put(ft + "leassee", ownerStats.get(ft + "leassee") + (agent.isLeaser() ? 1 : 0));
			ownerStats.put(ft + "income", ownerStats.get(ft + "income") + agent.getIncome());
			ownerStats.put(ft + "wFinancial", ownerStats.get(ft + "wFinancial") + agent.getNeedWeights().get(NeedWeight.COSTS));
			ownerStats.put(ft + "wFunctional", ownerStats.get(ft + "wFunctional") + agent.getNeedWeights().get(NeedWeight.FUNCTIONALITY));
			ownerStats.put(ft + "wSocial", ownerStats.get(ft + "wSocial") + agent.getNeedWeights().get(NeedWeight.SOCIAL));
			ownerStats.put(ft + "wEnvironmental", ownerStats.get(ft + "wEnvironmental") + agent.getNeedWeights().get(NeedWeight.PREFERENCES));
			
			ownerStats.put(ft + "ExpSatisfaction", ownerStats.get(ft + "ExpSatisfaction") + agent.getSatisfaction());
			ownerStats.put(ft + "ExpSatisCost", ownerStats.get(ft + "ExpSatisCost") + agent.getCostsSatisfaction());
			ownerStats.put(ft + "ExpSatisFunc", ownerStats.get(ft + "ExpSatisFunc") + agent.getFunctionalitySatisfaction());
			ownerStats.put(ft + "ExpSatisSoc", ownerStats.get(ft + "ExpSatisSoc") + agent.getSocialSatisfaction());
			ownerStats.put(ft + "ExpSatisPref", ownerStats.get(ft + "ExpSatisPref") + agent.getPreferenceSatisfaction());
			ownerStats.put(ft + "ExpUncertainty", ownerStats.get(ft + "ExpUncertainty") + agent.getUncertainty());
			ownerStats.put(ft + "ExpUncerCost", ownerStats.get(ft + "ExpUncerCost") + agent.getCostsUncertainty());
			ownerStats.put(ft + "ExpUncerFunc", ownerStats.get(ft + "ExpUncerFunc") + agent.getFunctionalityUncertainty());
			ownerStats.put(ft + "ExpUncerSoc", ownerStats.get(ft + "ExpUncerSoc") + agent.getSocialUncertainty());
			ownerStats.put(ft + "ExpUncerPref", ownerStats.get(ft + "ExpUncerPref") + agent.getPreferenceUncertainty());
			
			ownerStats.put(ft + agent.getLastStrategy().toString(), ownerStats.get(ft + agent.getLastStrategy().toString()) + 1);


			/* Determine availability of fuel types by aggregating over all car models
			Significantly increases computation costs. Only use when needed. */
//			for (Vehicle vehicle : Marketplace.getNewAndEVOccasions())
//			{
//				if (vehicle.getFuelType() != FuelType.GASOLINE && agent.modelWithinPriceRange(vehicle.getModel()))
//				{
//					// Proportion of agents familiar with/able to purchase EV 
//					if (vehicle.getFuelType() == FuelType.ELECTRIC && !knowsEV)
//					{
//						if (agent.isLeaser())
//							ownerStats.put("knows"+FuelType.ELECTRIC+"lease", ownerStats.get("knows"+FuelType.ELECTRIC+"lease")+1);
//						totalKnowsEV++;
//						knowsEV = true;
//					}
//					if (vehicle.getFuelType() == FuelType.HYBRID && !knowsPHEV)
//					{
//						if (agent.isLeaser())
//							ownerStats.put("knows"+FuelType.HYBRID+"lease", ownerStats.get("knows"+FuelType.HYBRID+"lease")+1);
//						totalKnowsPHEV++;
//						knowsPHEV = true;
//					}
//				}
//				if (knowsEV && knowsPHEV)
//					break;	
//			}		
			Map<Vehicle, Float> vehiclesWithYearlyCosts = new HashMap<Vehicle, Float>();
			float cheapestYearlyCosts = -1;
			for (Vehicle vehicle : Marketplace.getAllModels())
			{
				// Car within price range, available on market and known by agent?
				if (agent.canBuyModel(vehicle.getModel()))
				{
					// Yearly costs
					float yearlyCosts = agent.computeYearlyCosts(vehicle.getModel(), true);
					vehiclesWithYearlyCosts.put(vehicle, yearlyCosts);
					if (yearlyCosts < cheapestYearlyCosts || cheapestYearlyCosts == -1)
						cheapestYearlyCosts = yearlyCosts;
					// Count
					knownCount.put(vehicle.getFuelType().toString(), knownCount.get(vehicle.getFuelType().toString())+1);
				}
			}
			// Determine perceived satisfaction
			for (Entry<Vehicle, Float> entry : vehiclesWithYearlyCosts.entrySet())
			{
				Vehicle vehicle = entry.getKey();
				
				float costs_satisfaction = Math.max(0, 1 - ((entry.getValue() - cheapestYearlyCosts) / cheapestYearlyCosts));
				float functionality_satisfaction = agent.computeFunctionalitySatisfaction(vehicle.getModel());
				ownerStats.put(vehicle.getFuelType().toString() + "PercSatisCost", ownerStats.get(vehicle.getFuelType().toString() + "PercSatisCost")+costs_satisfaction);
				ownerStats.put(vehicle.getFuelType().toString() + "PercSatisFunc", ownerStats.get(vehicle.getFuelType().toString() + "PercSatisFunc")+functionality_satisfaction);
				ownerStats.put(vehicle.getFuelType().toString() + "PercSatisSoc", ownerStats.get(vehicle.getFuelType().toString() + "PercSatisSoc")+agent.computeSocialSatisfaction(vehicle.getModel(), (float)(costs_satisfaction + functionality_satisfaction)/2));
				ownerStats.put(vehicle.getFuelType().toString() + "PercSatisPref", ownerStats.get(vehicle.getFuelType().toString() + "PercSatisPref")+agent.computePreferenceSatisfaction(vehicle.getModel()));
				ownerStats.put(vehicle.getFuelType().toString() + "PercSatisfaction", ownerStats.get(vehicle.getFuelType().toString() + "PercSatisfaction")+agent.computePerceivedNeedSatisfaction(vehicle.getModel(), costs_satisfaction));
				
				// Uncertainty
				ownerStats.put(vehicle.getFuelType().toString() + "PercUncertainty", ownerStats.get(vehicle.getFuelType().toString() + "PercUncertainty")+agent.computePerceivedNeedUncertainty(vehicle.getModel()));
				ownerStats.put(vehicle.getFuelType().toString() + "PercUncerCost", ownerStats.get(vehicle.getFuelType().toString() + "PercUncerCost")+agent.computePerceivedCostUncertainty(vehicle.getModel()));
				ownerStats.put(vehicle.getFuelType().toString() + "PercUncerFunc", ownerStats.get(vehicle.getFuelType().toString() + "PercUncerFunc")+agent.computePerceivedFuncUncertainty(vehicle.getModel()));
				ownerStats.put(vehicle.getFuelType().toString() + "PercUncerSoc", ownerStats.get(vehicle.getFuelType().toString() + "PercUncerSoc")+agent.computePerceivedSocialUncertainty(vehicle.getModel()));
				ownerStats.put(vehicle.getFuelType().toString() + "PercUncerPref", ownerStats.get(vehicle.getFuelType().toString() + "PercUncerPref")+agent.computePerceivedPreferenceUncertainty(vehicle.getModel()));
			}
		}
		
//		// Update easy-access characteristics
		ownerStats.put("Debt", debt / agents.size());
		ownerStats.put("Emissions", emissions / distance);		
		ownerStats.put("Age", vehicleAge / agents.size());
		ownerStats.put("Duration", ownDuration / agents.size());
		ownerStats.put("knows"+FuelType.ELECTRIC+"lease", ownerStats.get("knows"+FuelType.ELECTRIC+"lease")/totalKnowsEV);
		ownerStats.put("knows"+FuelType.HYBRID+"lease", ownerStats.get("knows"+FuelType.HYBRID+"lease")/totalKnowsPHEV);
		ownerStats.put("knows"+FuelType.ELECTRIC+"total", (double)totalKnowsEV / agents.size());
		ownerStats.put("knows"+FuelType.HYBRID+"total", (double)totalKnowsPHEV / agents.size());
		ownerStats.put("IsSatisfiedNewAgent", ownerStats.get("IsSatisfiedNewAgent")/buysNew);
		for (FuelType ft : FuelType.values())
		{		
			ownerStats.put(ft.toString() + "IsSatisfied", ownerStats.get(ft.toString() + "IsSatisfied")/ownerStats.get(ft.toString() + "Total"));
			ownerStats.put(ft.toString() + "Ambition", ownerStats.get(ft.toString() + "Ambition")/ownerStats.get(ft.toString() + "Total"));
			ownerStats.put(ft.toString() + "uncTolerance", ownerStats.get(ft.toString() + "uncTolerance")/ownerStats.get(ft.toString() + "Total"));

			ownerStats.put(ft.toString() + "TotalPerNewAgent", ownerStats.get(ft.toString() + "TotalPerNewAgent")/buysNew);
			ownerStats.put(ft.toString() + "TotalPerLeaseAgent", ownerStats.get(ft.toString() + "TotalPerLeaseAgent")/leasers);
			ownerStats.put(ft.toString() + "TotalPerOccAgent", ownerStats.get(ft.toString() + "TotalPerOccAgent")/occasioners);
			
			ownerStats.put(ft.toString() + "hasHomeCharge", ownerStats.get(ft.toString() + "hasHomeCharge")/ownerStats.get(ft.toString() + "Total"));
			ownerStats.put(ft.toString() + "hasWorkCharge", ownerStats.get(ft.toString() + "hasWorkCharge")/ownerStats.get(ft.toString() + "Total"));
			ownerStats.put(ft.toString()  + "YearlyKm", ownerStats.get(ft.toString()  + "YearlyKm")/ownerStats.get(ft.toString() + "Total"));
			ownerStats.put(ft.toString() + "leassee", ownerStats.get(ft.toString() + "leassee")/ownerStats.get(ft.toString() + "Total"));
			ownerStats.put(ft.toString() + "income", ownerStats.get(ft.toString() + "income")/ownerStats.get(ft.toString() + "Total"));
			ownerStats.put(ft.toString() + "wFinancial", ownerStats.get(ft.toString() + "wFinancial")/ownerStats.get(ft.toString() + "Total"));
			ownerStats.put(ft.toString() + "wFunctional", ownerStats.get(ft.toString() + "wFunctional")/ownerStats.get(ft.toString() + "Total"));
			ownerStats.put(ft.toString() + "wSocial", ownerStats.get(ft.toString() + "wSocial")/ownerStats.get(ft.toString() + "Total"));
			ownerStats.put(ft.toString() + "wEnvironmental", ownerStats.get(ft.toString() + "wEnvironmental")/ownerStats.get(ft.toString() + "Total"));


			// Update experiences
			ownerStats.put(ft.toString() + "ExpSatisfaction", ownerStats.get(ft.toString() + "ExpSatisfaction")/ownerStats.get(ft.toString() + "Total"));
			ownerStats.put(ft.toString() + "ExpSatisCost", ownerStats.get(ft.toString() + "ExpSatisCost")/ownerStats.get(ft.toString() + "Total"));
			ownerStats.put(ft.toString() + "ExpSatisFunc", ownerStats.get(ft.toString() + "ExpSatisFunc")/ownerStats.get(ft.toString() + "Total"));
			ownerStats.put(ft.toString() + "ExpSatisSoc", ownerStats.get(ft.toString() + "ExpSatisSoc")/ownerStats.get(ft.toString() + "Total"));
			ownerStats.put(ft.toString() + "ExpSatisPref", ownerStats.get(ft.toString() + "ExpSatisPref")/ownerStats.get(ft.toString() + "Total"));
			ownerStats.put(ft.toString() + "ExpUncertainty", ownerStats.get(ft.toString() + "ExpUncertainty")/ownerStats.get(ft.toString() + "Total"));
			ownerStats.put(ft.toString() + "ExpUncerCost", ownerStats.get(ft.toString() + "ExpUncerCost")/ownerStats.get(ft.toString() + "Total"));
			ownerStats.put(ft.toString() + "ExpUncerFunc", ownerStats.get(ft.toString() + "ExpUncerFunc")/ownerStats.get(ft.toString() + "Total"));
			ownerStats.put(ft.toString() + "ExpUncerSoc", ownerStats.get(ft.toString() + "ExpUncerSoc")/ownerStats.get(ft.toString() + "Total"));
			ownerStats.put(ft.toString() + "ExpUncerPref", ownerStats.get(ft.toString() + "ExpUncerPref")/ownerStats.get(ft.toString() + "Total"));
			
			ownerStats.put(ft.toString() + InformationStrategy.IMITATION.toString(), ownerStats.get(ft.toString() + InformationStrategy.IMITATION.toString()) / ownerStats.get(ft.toString() + "Total"));
			ownerStats.put(ft.toString() + InformationStrategy.INQUIRING.toString(), ownerStats.get(ft.toString() + InformationStrategy.INQUIRING.toString()) / ownerStats.get(ft.toString() + "Total"));
			ownerStats.put(ft.toString() + InformationStrategy.OPTIMISING.toString(), ownerStats.get(ft.toString() + InformationStrategy.OPTIMISING.toString()) / ownerStats.get(ft.toString() + "Total"));
			ownerStats.put(ft.toString() + InformationStrategy.REPETITION.toString(), ownerStats.get(ft.toString() + InformationStrategy.REPETITION.toString()) / ownerStats.get(ft.toString() + "Total"));


			// Update perception
			ownerStats.put(ft.toString() + "PercSatisfaction", ownerStats.get(ft.toString() + "PercSatisfaction")/knownCount.get(ft.toString()));
			ownerStats.put(ft.toString() + "PercUncertainty", ownerStats.get(ft.toString() + "PercUncertainty")/knownCount.get(ft.toString()));
			ownerStats.put(ft.toString() + "PercSatisFunc", ownerStats.get(ft.toString() + "PercSatisFunc")/knownCount.get(ft.toString()));
			ownerStats.put(ft.toString() + "PercSatisCost", ownerStats.get(ft.toString() + "PercSatisCost")/knownCount.get(ft.toString()));
			ownerStats.put(ft.toString() + "PercSatisSoc", ownerStats.get(ft.toString() + "PercSatisSoc")/knownCount.get(ft.toString()));
			ownerStats.put(ft.toString() + "PercSatisPref", ownerStats.get(ft.toString() + "PercSatisPref")/knownCount.get(ft.toString()));
			ownerStats.put(ft.toString() + "PercUncerFunc", ownerStats.get(ft.toString() + "PercUncerFunc")/knownCount.get(ft.toString()));
			ownerStats.put(ft.toString() + "PercUncerCost", ownerStats.get(ft.toString() + "PercUncerCost")/knownCount.get(ft.toString()));
			ownerStats.put(ft.toString() + "PercUncerSoc", ownerStats.get(ft.toString() + "PercUncerSoc")/knownCount.get(ft.toString()));
			ownerStats.put(ft.toString() + "PercUncerPref", ownerStats.get(ft.toString() + "PercUncerPref")/knownCount.get(ft.toString()));
			
			// Transform fuel technology count into proportion
			ownerStats.put(ft.toString() + "Total", ownerStats.get(ft.toString() + "Total")/agents.size());
		}

		return ownerStats;
	}

	
	public void writeToFile(File file, String label, String value)
	{
		try {
			FileWriter fw = new FileWriter(file.getAbsoluteFile(), true);
			BufferedWriter bw = new BufferedWriter(fw);
			if (RepastEssentials.GetTickCount() == 1)
				bw.write("\n" + new Date() + ";" + label + ";");
			bw.write(value + ";");
			bw.flush();
			bw.close();
		} catch (IOException e) {
			e.printStackTrace();
			System.out.println("Problem writing to file: " + file.getAbsolutePath());
		}	
	}
	
	public File createFile(String filename)
	{
		File file = new File(filename);
		// if file doesnt exists, then create it
		if (!file.exists()) {
			try {
				file.createNewFile();
				FileWriter fw = new FileWriter(file.getAbsoluteFile(), true);
				BufferedWriter bw = new BufferedWriter(fw);
				bw.write(";;");
				for (int i = 1; i < Scenarios.getDuration(); ++i)
					bw.write(i + ";");
				bw.flush();
				bw.close();
			} catch (IOException e) {
				e.printStackTrace();
				System.out.println("Cannot create file: " + filename);
			}
		}
		return file;
	}
	
	public void writeToFileFTs(Map<FuelType, File> files, String folder, String label, Map<String, Double> ownerStats)
	{
		for (FuelType ft : FuelType.values())
		{
			if (!files.containsKey(ft))
			{
				File file = createFile(d_file_path + folder + "/" + label +  ft + ".csv");
				files.put(ft, file);
			}
			File file = files.get(ft);
			try {
				FileWriter fw = new FileWriter(file.getAbsoluteFile(), true);
				BufferedWriter bw = new BufferedWriter(fw);
				if (RepastEssentials.GetTickCount() == 1)
					bw.write("\n" + new Date() + ";" + label + ";");
				bw.write(Double.toString(ownerStats.get(ft.toString()+label)).replace('.', ',') + ";");
				bw.flush();
				bw.close();
			} catch (IOException e) {
				e.printStackTrace();
				System.out.println("Problem writing to file: " + file.getAbsolutePath());
			}	
			
		}
	}
	
	public void writeToFileStrategies(Map<InformationStrategy, File> files, String folder, Map<String, Double> ownerStats)
	{
		for (InformationStrategy strategy : InformationStrategy.values())
		{
			if (!files.containsKey(strategy))
			{
				File file = createFile(d_file_path + folder + "/" + strategy + ".csv");
				files.put(strategy, file);
			}
			File file = files.get(strategy);
			try {
				FileWriter fw = new FileWriter(file.getAbsoluteFile(), true);
				BufferedWriter bw = new BufferedWriter(fw);
				if (RepastEssentials.GetTickCount() == 1)
					bw.write("\n" + new Date() + ";" + strategy + ";");
				bw.write(Double.toString(ownerStats.get(strategy.toString())).replace('.', ',') + ";");
				bw.flush();
				bw.close();
			} catch (IOException e) {
				e.printStackTrace();
				System.out.println("Problem writing to file: " + file.getAbsolutePath());
			}	
			
		}
	}

}
