/**
 * 
 */
package electric_vehicles.VehicleInfrastructure;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import repast.simphony.engine.schedule.ScheduledMethod;
import repast.simphony.essentials.RepastEssentials;
import repast.simphony.random.RandomHelper;
import electric_vehicles.Enums.FuelType;
import electric_vehicles.Enums.MarketType;
import electric_vehicles.Vehicles.Vehicle;

/**
 * @author Ayla
 *
 */
public class Marketplace {
	
	private static int d_total_agents;
	
	private static int d_newSales, d_leaseSales, d_occasionSales;
	private static double d_newPrice, d_leasePrice;
	private static int d_newEVs, d_newHybrids, d_newGas;
	private static int d_EVOccCount, d_HybridOccCount, d_GasOccCount;
	
	private static List<Float> d_occValueEV, d_occValuePHEV, d_occValueGAS;
	private static List<Float> d_occAgeEV, d_occAgePHEV, d_occAgeGAS;

	private static int d_scrap_age, d_totalloss;
	
	private static Map<String, Integer> registered_vehicles;
	private static Map<String, Integer> model_sales;
	private static Map<String, List<String>> model_upgrades;
	
	private static List<Vehicle> allModels, d_newAndEVOccasions;
	private static Map<String, Vehicle> leaseMarket;	
	private static Map<String, Vehicle> newMarket;
	private static Map<String, List<Vehicle> > occasionMarket;	
	
	private static float d_averageGasEmissions;
	
	// Constructor
	public Marketplace()
	{
		d_total_agents = 0;
		
		d_EVOccCount = 0;
		d_HybridOccCount = 0;
		d_GasOccCount = 0;
		
		d_newSales = 0;
		d_leaseSales = 0;
		d_occasionSales = 0;
		
		d_newPrice = 0;
		d_leasePrice = 0;
		
		d_newGas = 0;
		d_newEVs = 0;
		d_newHybrids = 0;
		
		d_EVOccCount = 0;
		d_HybridOccCount = 0;
		d_GasOccCount = 0;
		
		d_scrap_age = 0;
		d_totalloss = 0;
		
		d_occValueEV 		= new LinkedList<Float>();
		d_occValuePHEV 		= new LinkedList<Float>();
		d_occValueGAS 		= new LinkedList<Float>();
		d_occAgeEV 			= new LinkedList<Float>();
		d_occAgePHEV 		= new LinkedList<Float>();
		d_occAgeGAS 		= new LinkedList<Float>();
		
		registered_vehicles = new HashMap<String, Integer>();
		model_sales 		= new HashMap<String, Integer>();
		model_upgrades		= new HashMap<String, List<String>>();

		allModels 			= new LinkedList<Vehicle>();
		d_newAndEVOccasions	= new LinkedList<Vehicle>();
		leaseMarket 		= new HashMap<String, Vehicle>();		
		newMarket 			= new HashMap<String, Vehicle>();
		occasionMarket		= new HashMap<String, List<Vehicle> >() ;
		
		d_averageGasEmissions = 133;		
	}
	
	@ScheduledMethod(start = 52, interval = 52, priority = 1)
	public void yearlyStats()
	{
		//System.out.println("Total new sales: " + d_newSales);
		//System.out.println("Total occasion sales: " + d_occasionSales);
	}
	
	
	@ScheduledMethod(start = 0 , interval = 1, priority = 1)
	public void maintainMarket()
	{
		// Compute average gasoline emissions
		int gasCount = 0;
		int emissions = 0;
		for (Vehicle vehicle : allModels)
		{
			if (vehicle.getFuelType() == FuelType.GASOLINE)
			{
				gasCount++;
				emissions += vehicle.getEmissions();
			}
		}
		d_averageGasEmissions = (float)emissions/gasCount;
		
		// Analyze occasion market
		d_EVOccCount = 0;
		d_HybridOccCount = 0;
		d_GasOccCount = 0;
		
		List<Float> electricValues = new LinkedList<Float>();
		List<Float> hybridValues = new LinkedList<Float>();
		List<Float> gasolineValues = new LinkedList<Float>();
		List<Float> electricAges = new LinkedList<Float>();
		List<Float> hybridAges = new LinkedList<Float>();
		List<Float> gasolineAges = new LinkedList<Float>();
		
		int currentTick = (int)RepastEssentials.GetTickCount();
		
		// Store info for statistics
		for (List<Vehicle> vehicles : occasionMarket.values())
		{
			if (!vehicles.isEmpty())
			{			
				switch(vehicles.get(0).getFuelType())
				{
				case ELECTRIC:	d_EVOccCount += vehicles.size();
								for(Vehicle vehicle : vehicles) {
									electricValues.add(vehicle.getValue());
									electricAges.add((currentTick - vehicle.getFirstPurchaseTick())/52);
								}
								break;
				case GASOLINE:	d_GasOccCount += vehicles.size();
								for(Vehicle vehicle : vehicles) {
									gasolineValues.add(vehicle.getValue());
									gasolineAges.add((currentTick - vehicle.getFirstPurchaseTick())/52);
								}
								break;
				case HYBRID:	d_HybridOccCount += vehicles.size();
								for(Vehicle vehicle : vehicles) {
									hybridValues.add(vehicle.getValue());
									hybridAges.add((currentTick - vehicle.getFirstPurchaseTick())/52);
								}
								break;
				}
			}
		}
		int total_occasions = Math.max((int)(d_total_agents * 0.04), 1);
				
		// Maintain stability of occasion market
		for (Entry<String,List<Vehicle>> set : occasionMarket.entrySet())
		{
			// Don't have more than three times the registered occasions available on the market.
			int surplus = (int) ( set.getValue().size() - ( ((float)registered_vehicles.get(set.getKey()) / d_total_agents) * 2 * total_occasions ) ); 
			for (int i = 2; i < surplus; ++ i)
				occasionMarket.get(set.getKey()).remove(0);
		}
		
		// Store prices and ages of occasion vehicles used for statistics
		d_occValueEV = electricValues;
		d_occValuePHEV = hybridValues;
		d_occValueGAS = gasolineValues;
		d_occAgeEV = electricAges;
		d_occAgePHEV = hybridAges;
		d_occAgeGAS = gasolineAges;
		Collections.sort(d_occValueEV);
		Collections.sort(d_occValuePHEV);
		Collections.sort(d_occValueGAS);
		Collections.sort(d_occAgeEV);
		Collections.sort(d_occAgePHEV);
		Collections.sort(d_occAgeGAS);

	}
	
	public static List<Vehicle> getNewAndEVOccasions()
	{
		return d_newAndEVOccasions;
	}
	
	public static float getAverageGasolineEmissions()
	{
		return d_averageGasEmissions;
	}

	public static Vehicle getLatestVersion(String model)
	{
		String model_series = model.substring(0, model.indexOf('_'));
		for (Vehicle vehicle: newMarket.values())
		{
			if (vehicle.getModelSeries().equals(model_series))
				return vehicle;
		}
		System.out.println("ERROR: Latest version of " + model + " not found in new market.");
		return null;
	}
	
	/* Introduce a new car type to be bought or leased
	 */
	public static void introduceCar(Vehicle vehicle)
	{
		
		//System.out.println("Introduced new vehicle: " + vehicle.getModel() + " of series " + vehicle.getModelSeries());
		model_sales.put(vehicle.getModel(), 0);
		registered_vehicles.put(vehicle.getModel(), 0);
		if (!model_upgrades.containsKey(vehicle.getModelSeries()))
			model_upgrades.put(vehicle.getModelSeries(), new LinkedList<String>());
		if (!model_upgrades.get(vehicle.getModelSeries()).contains(vehicle.getModel()))
			model_upgrades.get(vehicle.getModelSeries()).add(vehicle.getModel());
		
		allModels.add(vehicle);
		leaseMarket.put(vehicle.getModel(), vehicle);		
		newMarket.put(vehicle.getModel(), vehicle);
		occasionMarket.put(vehicle.getModel(), new LinkedList<Vehicle>());
	}
	
	public static void removeModel(String model)
	{
		newMarket.remove(model);
		leaseMarket.remove(model);
		for (int i = 0; i < allModels.size(); ++i)
		{
			if (allModels.get(i).getModel().equals(model))
			{
				allModels.remove(i);
				return;
			}
		}
		System.out.println("ERROR: Model to be removed not found in allModels list");
	}
	
	public static int getUpgradesSinceThisVehicle(String model) 
	{
		String model_series = model.substring(0, model.indexOf('_'));
		int this_version = model_upgrades.get(model_series).indexOf(model);
		int newest_version = model_upgrades.get(model_series).size() - 1;
		return newest_version - this_version;
	}
	
	/* Returns the new market (i.e. all available models) */
	public static List<Vehicle> getAllModels()							{ return allModels; }
	
	/* Return the new price of a certain car model */
	public static float getNewPrice(String model)						{ return newMarket.containsKey(model) ? newMarket.get(model).getPrice() : -1; }
	
	/* Return the cars in the occasion market so agent can choose a car to buy. */
	public static List<Vehicle> getOccasionsFromModel(String model)		{ return occasionMarket.get(model); }
	
	/* Return whether an occasion model is available within specific price range*/
	public static boolean occasionMarketHasModelInPriceRange(String model, List<Float> price_range)
	{
		for (Vehicle vehicle : occasionMarket.get(model))
			if (vehicle.getValue() >= price_range.get(0) && vehicle.getValue() <= price_range.get(1))
				return true;
		return false;
	}	
	
	/* Return price of cheapest available occasion model within specific price range*/
	public static float cheapestVehicleInOccasionMarketInPriceRange(String model, List<Float> price_range)
	{
		float cheapest_vehicle = 0;
		for (Vehicle vehicle : occasionMarket.get(model))
			if (vehicle.getValue() >= price_range.get(0) && vehicle.getValue() <= price_range.get(1) && vehicle.getValue() > cheapest_vehicle)
				cheapest_vehicle = vehicle.getValue();
		return cheapest_vehicle;
	}
	
	/* Adds a new vehicle to the occasion market. Randomly decides the km count and age of the vehicle. */
	public static void addNewVehicleToOccasionMarket(String model)
	{
		Vehicle occasion = newMarket.get(model).copy();
		occasion.setKmCount(RandomHelper.nextIntFromTo(10000, 150000));
		occasion.setIntroductionTick((int)RepastEssentials.GetTickCount() - RandomHelper.nextIntFromTo(26, 520));
		occasion.setFirstPurchaseTick(occasion.getIntroductionTick());

		occasionMarket.get(model).add(occasion);	
	}
	
	/* Lets agent purchase a vehicle. In case of 'new' or 'lease', 
	 * there is unlimited supply of each model. In case of 'occasion',
	 * the purchased vehicle is removed from the occasion market.
	 */
	public static Vehicle buyNewCar(String model)
	{
		d_newSales++;
		Vehicle vehicle = newMarket.get(model).copy();
		registered_vehicles.put(vehicle.getModel(), registered_vehicles.get(vehicle.getModel())+1);
		d_newPrice += vehicle.getPrice();
		switch (vehicle.getFuelType())
		{
		case ELECTRIC:	d_newEVs++;
						break;
		case GASOLINE:	d_newGas++;
						break;
		case HYBRID:	d_newHybrids++;
						break;		
		}
		return vehicle;
	}
	
	public static Vehicle leaseNewCar(String model)
	{
		d_leaseSales++;
		Vehicle vehicle = leaseMarket.get(model).copy();
		registered_vehicles.put(vehicle.getModel(), registered_vehicles.get(vehicle.getModel())+1);
		d_leasePrice += vehicle.getPrice();
		switch (vehicle.getFuelType())
		{
		case ELECTRIC:	d_newEVs++;
						break;
		case GASOLINE:	d_newGas++;
						break;
		case HYBRID:	d_newHybrids++;
						break;		
		}
		return vehicle;
	}
	
	public static Vehicle buyOccasionCar(String model, int index)
	{
		d_occasionSales++;
		Vehicle vehicle = occasionMarket.get(model).remove(index);
		vehicle.resetPurchaseTick();
		registered_vehicles.put(vehicle.getModel(), registered_vehicles.get(vehicle.getModel())+1);
		return vehicle;
	}
	
	/* Allows an agent to sell its current vehicle so it becomes available
	 * on the occasion market. Randomly decide whether car is still fit enough
	 * for the occasion market.
	 */
	public static void sellCar(Vehicle vehicle, float netto_value)
	{
		registered_vehicles.put(vehicle.getModel(), registered_vehicles.get(vehicle.getModel())-1);
		// If vehicle is not total loss, reset failure costs, replenish energy and put car on occasion market
		if (netto_value > 0)
		{
			vehicle.resetTotalFailureCosts();
			vehicle.resetEnergy();
			occasionMarket.get(vehicle.getModel()).add(vehicle);
			return;
		}
		d_scrap_age += RepastEssentials.GetTickCount() - vehicle.getFirstPurchaseTick();
		d_totalloss++;
	}
	
	/* Saturates the occasion market with a number of cars at 
	 * the beginning of the simulation
	 */
	public static void initialiseOccasionMarket(int total_agents)
	{
		d_total_agents = total_agents;
		int total_occasions = Math.max((int)(total_agents * 0.0387), 1);
		
		for (Vehicle vehicle : newMarket.values())
		{
			int sale_proportion = (int)Math.ceil(((float)model_sales.get(vehicle.getModel())/total_agents) * total_occasions);
			for (int i = 0 ; i < sale_proportion; ++i)
				addNewVehicleToOccasionMarket(vehicle.getModel());
		}
	}
	
	/* Return the initial car that is assigned to an agent. Also set
	 * its km count to that which respondent indicated.
	 */
	public static Vehicle getInitialCar(FuelType type, int km_count, int current_vehicle_age, int vehicle_in_possesion_age, List<Float> new_price_range, boolean buys_new)
	{
		Vehicle new_vehicle = null;
		float min_price = (float) (buys_new? new_price_range.get(0) : new_price_range.get(0) / Math.pow(0.85f, (current_vehicle_age - vehicle_in_possesion_age)));
		float max_price = (float) (buys_new? new_price_range.get(1) : new_price_range.get(1) / Math.pow(0.85f, (current_vehicle_age - vehicle_in_possesion_age)));
		
		for (Vehicle vehicle : newMarket.values())
			if(vehicle.getFuelType() == type && vehicle.getPrice() >= min_price && vehicle.getPrice() <= max_price)
			{
				new_vehicle = vehicle.copy();
				new_vehicle.setIntroductionTick(Math.min(0, -current_vehicle_age * 52 + RandomHelper.nextIntFromTo(-26, 26)));
				new_vehicle.setFirstPurchaseTick(new_vehicle.getIntroductionTick());
				new_vehicle.setKmCount(km_count * (Math.abs(new_vehicle.getIntroductionTick())/52));
				new_vehicle.setPurchaseTick(Math.min(0, -vehicle_in_possesion_age * 52 + RandomHelper.nextIntFromTo(-26, 26)));
				model_sales.put(new_vehicle.getModel(), model_sales.get(new_vehicle.getModel())+1);		// Increase sales of this model
				registered_vehicles.put(new_vehicle.getModel(), registered_vehicles.get(new_vehicle.getModel())+1);	
				break;
			}
		if (new_vehicle == null)
		{
			float smallest_difference = 0;
			String closest_model = "";
			for (Vehicle vehicle : newMarket.values())
			{
				float difference = (float) Math.min(Math.abs(vehicle.getPrice() - min_price), Math.abs(vehicle.getPrice() - max_price));
				if(vehicle.getFuelType() == type && (difference < smallest_difference || smallest_difference == 0))
				{
					smallest_difference = difference;
					closest_model = vehicle.getModel();
				}
			}
			if (closest_model == "")
				System.out.println("Warning: Cannot find suitable fuel type in market to initialise agent's car. Price range is: " + new_price_range + "Fuel type is " + type);
			new_vehicle = newMarket.get(closest_model).copy();
			new_vehicle.setKmCount(km_count * (Math.abs(new_vehicle.getIntroductionTick())/52));
			new_vehicle.setIntroductionTick(Math.min(0, -current_vehicle_age * 52 + RandomHelper.nextIntFromTo(-26, 26)));
			new_vehicle.setFirstPurchaseTick(new_vehicle.getIntroductionTick());
			new_vehicle.setPurchaseTick(Math.min(0, -vehicle_in_possesion_age * 52 + RandomHelper.nextIntFromTo(-26, 26)));
			model_sales.put(new_vehicle.getModel(), model_sales.get(new_vehicle.getModel())+1);		// Increase sales of this model
			registered_vehicles.put(new_vehicle.getModel(), registered_vehicles.get(new_vehicle.getModel())+1);	
		}
		return new_vehicle;
	}
	
	/* For printing all the vehicles in a specific market
	 */
	public static void printMarket(MarketType marketType)
	{
		System.out.println("\nContents of Market: " + marketType);
		Collection<Vehicle> market = null;
		switch(marketType)
		{
			case NEW: 		market = newMarket.values();
							break;
			case LEASE : 	market = leaseMarket.values();
							break;
			case OCCASION:	market = new LinkedList<Vehicle>();
							for (List<Vehicle> list : occasionMarket.values()) 	
								market.addAll(list);
		}
		
		if (market == null || market.size() == 0)
		{
			System.out.println("WARNING: Cannot print empty market");
			return;
		}
		System.out.println("Fuel type:\tModel:\tPrice:\t\tKM costs:\tRange:\tEnv. Impact:\tKm Count:\tEnergy:\tValue:\t\t\tFailure costs:");		
		for(Vehicle vehicle : market)
			System.out.println(vehicle.printVehicle());
	}

	
	/* 
	 * *************************************************** STATISTICS FUNCTIONS ****************************************************
	 * 
	*/	
	
	static public int EVsOccasionsCount()
	{
		return d_EVOccCount;
	}
	
	static public int HybridOccasionsCount()
	{
		return d_HybridOccCount;
	}
	
	static public int GasolineOccasionsCount()
	{
		return d_GasOccCount;
	}
	
	static public int getOccasionSales()
	{
		return d_occasionSales;
	}
	
	static public int getNewSales()
	{
		return d_newSales + d_leaseSales;
	}
	
	static public float getMedianElectricOccValue()
	{
		if (d_occValueEV.size() == 0) {
			return 0;
		}
		if (d_occValueEV.size() == 1) {
			return d_occValueEV.get(0);
		}
		return d_occValueEV.get((d_occValueEV.size()/2)-1);
	}

	static public float getMedianHybridOccValue()
	{
		if (d_occValuePHEV.size() == 0) {
			return 0;
		}
		if (d_occValuePHEV.size() == 1) {
			return d_occValuePHEV.get(0);
		}
		return d_occValuePHEV.get((d_occValuePHEV.size()/2)-1);
	}
	
	static public float getMedianGasolineOccValue()
	{
		if (d_occValueGAS.size() == 0) {
			return 0;
		}
		if (d_occValueGAS.size() == 1) {
			return d_occValueGAS.get(0);
		}
		return d_occValueGAS.get((d_occValueGAS.size()/2)-1);
	}
	
	static public float getMeanElectricOccValue()
	{
		double totalValue = 0;
		for (Float value : d_occValueEV) {
			totalValue += value;
		}
		return d_occValueEV.size() == 0 ? 0 : (float)totalValue / d_occValueEV.size();
	}
	
	static public float getMeanHybridOccValue()
	{
		double totalValue = 0;
		for (Float value : d_occValuePHEV) {
			totalValue += value;
		}
		return d_occValuePHEV.size() == 0 ? 0 : (float)totalValue / d_occValuePHEV.size();
	}
	
	static public float getMeanGasolineOccValue()
	{
		double totalValue = 0;
		for (Float value : d_occValueGAS) {
			totalValue += value;
		}
		return d_occValueGAS.size() == 0 ? 0 : (float)totalValue / d_occValueGAS.size();
	}
	
	static public float getMeanElectricOccAge()
	{
		double totalValue = 0;
		for (Float value : d_occAgeEV) {
			totalValue += value;
		}
		return d_occAgeEV.size() == 0 ? 0 : (float)totalValue / d_occAgeEV.size();
	}
	
	static public float getMeanHybridOccAge()
	{
		double totalValue = 0;
		for (Float value : d_occAgePHEV) {
			totalValue += value;
		}
		return d_occAgePHEV.size() == 0 ? 0 : (float)totalValue / d_occAgePHEV.size();
	}
	
	static public float getMeanGasolineOccAge()
	{
		double totalValue = 0;
		for (Float value : d_occAgeGAS) {
			totalValue += value;
		}
		return d_occAgeGAS.size() == 0 ? 0 : (float)totalValue / d_occAgeGAS.size();
	}
	
	public static List<Float> getAllElectricOccValues() { return d_occValueEV; }	
	public static List<Float> getAllHybridOccValues() 	{ return d_occValuePHEV; }	
	public static List<Float> getAllGasolineOccValues() { return d_occValueGAS; }
	public static List<Float> getAllElectricOccAges() 	{ return d_occAgeEV; }	
	public static List<Float> getAllHybridOccAges() 	{ return d_occAgePHEV; }	
	public static List<Float> getAllGasolineOccAges() 	{ return d_occAgeGAS; }
	
	static public float getProportionGasSales()
	{
		return d_newGas / (float)(d_newGas + d_newEVs + d_newHybrids);
	}
	
	static public float getProportionEVSales()
	{
		return d_newEVs / (float)(d_newGas + d_newEVs + d_newHybrids);
	}
	
	static public float getProportionPHEVSales()
	{
		return d_newHybrids / (float)(d_newGas + d_newEVs + d_newHybrids);
	}
	
	static public int getTotalLoss()
	{
		return d_totalloss;
	}
	
	static public float averageScrapAge(){
		return (float)d_scrap_age / d_totalloss / 52;
	}
	
	static public float averageNewPrice() {
		return (float) (d_newPrice / d_newSales);
	}
	
	static public float averageLeasePrice() {
		return (float) (d_leasePrice / d_leaseSales);
	}
	
	static public void resetYear()
	{
		d_newSales = 0;
		d_leaseSales = 0;
		d_occasionSales = 0;
		d_totalloss = 0;
		d_scrap_age = 0;
	}
	
	static public void resetQuarter()
	{
		d_newGas = 0;
		d_newEVs = 0;
		d_newHybrids = 0;
	}
	
	
}
