/**
 * 
 */
package electric_vehicles.VehicleInfrastructure;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.parameter.Parameters;
import electric_vehicles.Enums.FuelType;

/**
 * @author Ayla
 *
 */
public class Mechanic {
	
	// Per fuel type, the probability of encountering failures, compared to gasoline vehicle.
	static private Map<FuelType, Float> d_fail_ratios;
	
	// Per fuel type, the probability (Float) of different failure costs (Integer) per 50 km 
	static private  Map<FuelType, Map<Integer, Float>> d_failure_costs;

	
	public Mechanic()
	{
		d_failure_costs = new HashMap<FuelType, Map<Integer, Float>>();
		
		Parameters params = RunEnvironment.getInstance().getParameters();
		
		d_fail_ratios = new HashMap<FuelType, Float>();
		d_fail_ratios.put(FuelType.GASOLINE, 1.0f);
		d_fail_ratios.put(FuelType.ELECTRIC, ( Float )params.getValue("electric_fail_ratio"));
		d_fail_ratios.put(FuelType.HYBRID, ( Float )params.getValue("hybrid_fail_ratio"));		
		
		for (FuelType type : FuelType.values())
		{
			d_failure_costs.put(type, new HashMap<Integer, Float>());
			d_failure_costs.get(type).put(50, 0.01f * d_fail_ratios.get(type));
			d_failure_costs.get(type).put(500, 0.001f * d_fail_ratios.get(type));
			d_failure_costs.get(type).put(2000, 0.0005f * d_fail_ratios.get(type));
		}

	}
	
	static public float getSmallFailureProb(FuelType type)
	{
		return d_failure_costs.get(type).get(50);
	}
	
	static public float getMediumFailureProb(FuelType type)
	{
		return d_failure_costs.get(type).get(500);
	}
	
	static public float getLargeFailureProb(FuelType type)
	{
		return d_failure_costs.get(type).get(2000);
	}
	
	static public Map<Integer, Float> getFailureProbabilities(FuelType type)
	{
		return d_failure_costs.get(type);
	}
	
	static public float getAverageFailureCostsPerKm(FuelType type)
	{
		float costs = 0;
		for (Entry<Integer,Float> set : d_failure_costs.get(type).entrySet())
			costs += (set.getKey() * set.getValue());
		// convert to costs/km
		return costs/50;	
	}
	
	static public float getFailRatio(FuelType type)
	{
		return d_fail_ratios.get(type);
	}
	
	static public void adjustFailRatio(FuelType type, float adjustment)
	{
		float new_value = d_fail_ratios.get(type) + adjustment;
		d_fail_ratios.put(type, new_value);
		
		d_failure_costs.get(type).put(50, 0.01f * new_value);
		d_failure_costs.get(type).put(500, 0.001f * new_value);
		d_failure_costs.get(type).put(2000, 0.0005f * new_value);
	}
	
	/* 
	 * ****************************************************** Statistics *******************************************************
	 * 
	*/
	
	public float getElectricFailRatio()		{ return d_fail_ratios.get(FuelType.ELECTRIC); }
	public float getHybridFailRatio()		{ return d_fail_ratios.get(FuelType.HYBRID); }

	
	
	

}
