/**
 * 
 */
package electric_vehicles.VehicleInfrastructure;

import java.util.HashMap;
import java.util.Map;

import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.engine.schedule.ScheduledMethod;
/*import repast.simphony.engine.schedule.ScheduledMethod;
*/import repast.simphony.parameter.Parameters;
import electric_vehicles.Enums.ChargeLocation;
import electric_vehicles.Enums.ChargeOption;
import electric_vehicles.Enums.FuelType;

/**
 * @author Ayla
 *
 */
public class RechargeStation 
{
	static private int workcharge_agents;
	static private int homecharge_agents;
	static private  Map<ChargeLocation, Float> charge_probability;
	static private  Map<ChargeOption, Float> charge_time;
	static private Map<ChargeOption, Float> km_costs;
	static private int d_fastChargedEnergy;
	
	public RechargeStation()
	{
		Parameters params = RunEnvironment.getInstance().getParameters();
		
		charge_probability = new HashMap<ChargeLocation, Float>();
		charge_probability.put(ChargeLocation.HIGHWAY, ( Float )params.getValue("charge_probability"));
		charge_probability.put(ChargeLocation.HOME, 0.0f);
		charge_probability.put(ChargeLocation.WORK, ( Float )params.getValue("work_charge_probability"));
		workcharge_agents = 0;
		homecharge_agents = 0;
		
		charge_time = new HashMap<ChargeOption, Float>();
		charge_time.put(ChargeOption.SLOW, ( Float )params.getValue("electric_charge_time"));
		charge_time.put(ChargeOption.FAST, ( Float )params.getValue("electric_charge_time_fast"));
		charge_time.put(ChargeOption.GASOLINE, ( Float )params.getValue("gasoline_charge_time"));
		
		km_costs = new HashMap<ChargeOption, Float>();
		km_costs.put(ChargeOption.SLOW, ( Float )params.getValue("electric_km_costs"));
		km_costs.put(ChargeOption.FAST, ( Float )params.getValue("electric_km_costs_fast"));
		km_costs.put(ChargeOption.GASOLINE, ( Float )params.getValue("fuel_km_costs"));
	}
	
	public static void updateFastChargedEnergy(int energy) 	{ d_fastChargedEnergy += energy; }
	public static int getFastChargedEnergy()				{ return d_fastChargedEnergy; }
	public static void resetFastChargedEnergy()				{ d_fastChargedEnergy = 0; }
	public static void setHomeChargeProb(int total_agents)	{ charge_probability.put(ChargeLocation.HOME, (float)workcharge_agents/total_agents); }
	
	/* 
	 * ****************************************************** Charge probability *******************************************************
	 * 
	*/
	public static int getWorkChargeAgentsCount()			{ return workcharge_agents; }
	public static void increaseWorkChargeAgentsCount()		{ ++workcharge_agents; }
	public static int getHomeChargeAgentsCount()			{ return homecharge_agents; }
	public static void increaseHomeChargeAgentsCount()		{ ++homecharge_agents; }
	
	public static float getChargeProbability(ChargeLocation location)		
	{
		return charge_probability.get(location);
	}
	
	public static void adjustChargeProbability(ChargeLocation location, float adjustment)		
	{
		float new_value = charge_probability.get(location) + adjustment;
		charge_probability.put(location, new_value);
	}
	/* 
	 * ****************************************************** Charge time *******************************************************
	 * 
	*/
	public static float getChargeTime(ChargeOption option)
	{
		return charge_time.get(option);
	}
	
	// Overloaded function, used when we cannot hard code which fuel type we're dealing with (such as in the generic FuelTypeConcept class)
	public static float getChargeTime(FuelType type, ChargeOption option)
	{
		if (type == FuelType.ELECTRIC)
			return charge_time.get(option);
		return charge_time.get(ChargeOption.GASOLINE);
	}
	
	public static void adjustChargeTime(ChargeOption option, float adjustment)		
	{
		float new_value = charge_time.get(option) + adjustment;
		charge_time.put(option, new_value);
	}
	
	/* 
	 * ****************************************************** Charge costs *******************************************************
	 * 
	*/
	public static float getKmCosts(ChargeOption option)
	{
		return km_costs.get(option);
	}
	
	// Overloaded function, used when we cannot hard code which charge option we're dealing with (such as in the generic FuelTypeConcept class)
	public static float getKmCosts(FuelType type, ChargeOption option)
	{
		if (type == FuelType.GASOLINE)
			return km_costs.get(ChargeOption.GASOLINE);
		else if (type == FuelType.HYBRID)
			return (km_costs.get(ChargeOption.GASOLINE)+km_costs.get(ChargeOption.SLOW))/2;
		return km_costs.get(option);
	}
	
	public static void adjustKmCosts(ChargeOption option, float adjustment)		
	{
		float new_value = km_costs.get(option) + adjustment;
		km_costs.put(option, new_value);
	}
	
	/* 
	 * ****************************************************** Statistics *******************************************************
	 * 
	*/
	public float HomeChargeProb()	{ return charge_probability.get(ChargeLocation.HOME); }
	public float WorkChargeProb()	{ return charge_probability.get(ChargeLocation.WORK); }
	public float RoadChargeProb()	{ return charge_probability.get(ChargeLocation.HIGHWAY); }
	public float HomeChargeTime()	{ return charge_time.get(ChargeOption.FAST); }
	public float WorkChargeTime()	{ return charge_time.get(ChargeOption.SLOW); }
	public float RoadChargeTime()	{ return charge_time.get(ChargeOption.GASOLINE); }
	public float HomeChargeCosts()	{ return km_costs.get(ChargeOption.FAST); }
	public float WorkChargeCosts()	{ return km_costs.get(ChargeOption.SLOW); }
	public float RoadChargeCosts()	{ return km_costs.get(ChargeOption.GASOLINE); }
	
}
