package electric_vehicles.VehicleInfrastructure;

import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;

import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.parameter.Parameters;

public class Tax {
	
	static private int max_co2_emmissions;
	
	static private List<TaxLevel> bijtelling_levels;
	static private int max_bijtelling_level;
	
	static private List<TaxLevel> tax_levels;
	static private int max_tax_level;
	
	public Tax()
	{
		Parameters params = RunEnvironment.getInstance().getParameters();
			
		// Initialise bijtelling levels
		bijtelling_levels = new LinkedList<TaxLevel>();
		List<Integer> bijtelling_percentages = new LinkedList<Integer>(Arrays.asList(0,4,7,14,20,25));
		int min = 0;
		int max_level = 0;
		for (int level : bijtelling_percentages)
		{
			int max = ( Integer )params.getValue("bijtelling_level_" + level) + 1; 
			
			// Percentage not used?
			if (max - 1 < 0)
				continue;
			
			// Add tax level with specified emission boundaries
			bijtelling_levels.add(new TaxLevel(level, min, max));
			min = max;
			max_level = level;
		}
		max_co2_emmissions = min;
		max_bijtelling_level = max_level;
		
		// Initialize tax costs
		tax_levels = new LinkedList<TaxLevel>();
		List<Integer> tax_costs = new LinkedList<Integer>(Arrays.asList(0,20,50,100,150,200));
		min = 0;
		max_level = 0;
		for (int level : tax_costs)
		{
			int max = ( Integer )params.getValue("tax_level_" + level) + 1; 
			
			// Percentage not used?
			if (max - 1 < 0)
				continue;
			
			// Add tax level with specified emission boundaries
			tax_levels.add(new TaxLevel(level, min, max));
			min = max;
			max_level = level;
		}
		max_tax_level = max_level;
	}
	
	static public int getMaxEmmissions()				{ return max_co2_emmissions; }
	static public int getMaxBijtellingLevel()			{ return max_bijtelling_level; }
	static public int getMaxTaxLevel()					{ return max_tax_level; }
	
	/* Returns the bijtelling for a specific emission value 
	 */
	static public int getBijtellingLevel(int emission)
	{
		for (TaxLevel level : bijtelling_levels)
			if (level.isCorrectTaxLevel(emission))
				return level.getTax();
		
		// If emissions are higher than max emissions, pay max taxes.
		return max_bijtelling_level;
	}
	
	static public void adjustBijtellingLevels(List<TaxLevel> levels, int max_level, int max_emissions)
	{
		bijtelling_levels = levels;
		max_bijtelling_level = max_level;
		max_co2_emmissions = max_emissions;
	}

	
	/* Returns the tax costs for a specific emission value 
	 */
	static public int getTaxLevel(int emission)
	{
		for (TaxLevel level : tax_levels)
			if (level.isCorrectTaxLevel(emission))
				return level.getTax();
		
		// If emissions are higher than max emissions, pay max taxes.
		return max_tax_level;
	}
	
	static public void adjustTaxLevels(List<TaxLevel> levels, int max_level)
	{
		tax_levels = levels;
		max_tax_level = max_level;
	}
	
	static public float getIncomeTaxCategory(int income)	
	{ 
		if (income <= 2)
			return 33.1f;
		else if (income > 2 && income < 8)
			return 42;
		else
			return 52;
	}

}
