/**
 * 
 */
package electric_vehicles.Vehicles;


import java.util.Map;

import repast.simphony.essentials.RepastEssentials;
import electric_vehicles.Enums.ChargeOption;
import electric_vehicles.Enums.FuelType;
import electric_vehicles.VehicleInfrastructure.RechargeStation;

/**
 * @author Ayla
 *
 */
public class GasolineVehicle extends Vehicle {

	// Constructor when a new gasoline model is introduced
	public GasolineVehicle(FuelType fuelType, String model, int introductionTick, float price, int range, int emissions)
	{
		super(fuelType, model, introductionTick, price, range, emissions);
		d_model = d_model_series + "_" + emissions;
	}
	
	
	/* @see electric_vehicles.Vehicle#copy()
	 * Copy function when a gasoline model is bought (copied) from new or lease market
	 */
	public GasolineVehicle copy()
	{
		return new GasolineVehicle(d_fuelType, d_model_series, d_model, (int)RepastEssentials.GetTickCount(), d_price, d_range, d_emissions, d_failures);
	}
	
	// Constructor when a gasoline model is bought (copied) from new or lease market
	public GasolineVehicle(FuelType fuelType, String model_series, String model, int introductionTick, float price, int range, int emissions, Map<Integer, Float> failures)
	{
		super(fuelType, model_series, model, introductionTick, price, range, emissions, failures);
	}
	

	/* On top of Vehicle.getDriveLog(), this method determines specific attributes for a fuel car.
	 * @see electric_vehicles.Vehicle#getDriveLog(int)
	 */
	public DriveLog getDriveLog(int total_distance, boolean workTrip, boolean homeCharge, float recharge_moment)
	{
		DriveLog log = super.getDriveLog(total_distance, workTrip, homeCharge, recharge_moment);
		
		// When driving further than the current energy level of car allows, refill in between
		while (log.d_remainingKm > (int)(d_energy - d_range * recharge_moment) ) 
		{			
			log.d_remainingKm -= (int)(d_energy - d_range * recharge_moment);					// Drive as far as possible with current energy
			d_energy -= (int)(d_energy - d_range * recharge_moment);							// Reduce fuel in tank
			int newEnergy = refill();															// Refill tank and keep track of how often tank is refilled
			log.d_refillEnergy += newEnergy;													// Store amount of energy gained
			log.d_refillCosts += newEnergy * RechargeStation.getKmCosts(ChargeOption.GASOLINE);// Store costs of new energy
			log.d_minutesLate += RechargeStation.getChargeTime(ChargeOption.GASOLINE);			// Increase lateness
		}

		// When rest of the drive can be made without refilling, just subtract needed fuel from current energy
		if (log.d_remainingKm > 0)
			d_energy -= log.d_remainingKm;

		d_currentKmCosts = log.d_refillCosts == 0 ? d_currentKmCosts : log.d_refillCosts / log.d_refillEnergy;
		
		return log;
	}
	
	/* Refills the tank completely and returns the costs of the refill
	 */
	public int refill()
	{	
		int newEnergy = d_range - d_energy;
		d_energy += newEnergy;
		return newEnergy;		
	}

}
